# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

# TODO: Make all functions that take a cursor position return a tuple if the cursor
#       position is given, or just a single argument otherwise.

# TODO: To comply with bash, a backslash shouldn't have special meaning unless it's in
#       front of a special character, e.g. "\1" should be a backslash and 1.

# TODO: To comply with bash, backslashes shouldn't have special meaning inside single
#       quotes, e.g. '\' should be a single backslash while "\" would be a double quote
#       with the closing quote missing.


from collections import abc

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


DEFAULT_DELIMS = (' ',)
DEFAULT_ESCAPES = ('\\',)
DEFAULT_QUOTES = ('"', "'")


class Char(str):
    """Single character with parser-provided attributes"""

    def __new__(cls, char, *, string='', delim='', escape='', quote='',
                is_special=False, is_escaped=False, is_quoted=False):
        obj = super().__new__(cls, char)
        obj._string = string
        obj._delim = delim
        obj._escape = escape
        obj._quote = quote
        obj._is_special = bool(is_special)
        obj._is_escaped = bool(is_escaped)
        obj._is_quoted = bool(is_quoted)
        return obj

    @property
    def string(self):
        """
        Special delimiting/escaping/quoting string this character is a part of or
        empty string
        """
        return self._string

    @property
    def delim(self):
        """Delimiter string this character is a part of or empty string"""
        return self._delim

    @property
    def escape(self):
        """Escape string this character is a part of or is escaped with or empty string"""
        return self._escape

    @property
    def quote(self):
        """Quote string this character is a part of or quoted with or empty string"""
        return self._quote

    @property
    def is_special(self):
        """
        Whether this character is part of a special string and interpreted as such
        (i.e. not escaped or quoted)
        """
        return self._is_special

    @property
    def is_delim(self):
        """Whether this character is a delimiter"""
        return bool(self._delim)

    @property
    def is_escaped(self):
        """Whether this character is escaped (False for escapes)"""
        return self._is_escaped

    @property
    def is_quoted(self):
        """Whether this character comes after an opening quote (False for quotes)"""
        return self._is_quoted

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (str(self) == str(other) and
                    self._string == other._string and
                    self._delim == other._delim and
                    self._escape == other._escape and
                    self._quote == other._quote and
                    self._is_special == other._is_special and
                    self._is_escaped == other._is_escaped and
                    self._is_quoted == other._is_quoted)
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        kwargs = {}
        if self._string: kwargs['string'] = self._string
        if self._delim: kwargs['delim'] = self._delim
        if self._escape: kwargs['escape'] = self._escape
        if self._quote: kwargs['quote'] = self._quote
        if self._is_special: kwargs['is_special'] = self._is_special
        if self._is_escaped: kwargs['is_escaped'] = self._is_escaped
        if self._is_quoted: kwargs['is_quoted'] = self._is_quoted
        if kwargs:
            return '%s(%r, %s)' % (type(self).__name__, str(self),
                                   ', '.join('%s=%r' % (k,v) for k,v in kwargs.items()))
        else:
            return '%s(%r)' % (type(self).__name__, str(self))


def _on_any_substr(string, pos, substrs):
    """
    Whether the character at `string[pos]` is on a aubstring that exists in
    `substrs`

    Return the following tuple:
        (<bool: whether substring was found>,
         <str: the found substring or ''>,
         <int or None: position in `string` where substring was found or None>)
         <int or None: position in substring or None>)
    """
    if pos >= len(string):
        return False, '', None, None

    # Match the longest substrings before the shorter ones
    for substr in sorted(substrs, key=len, reverse=True):
        # Move position backwards until we find `substr`
        len_substr = len(substr)
        substr_pos = pos
        found_substr = ''
        for _ in range(len_substr):
            maybe_substr = string[substr_pos : substr_pos + len_substr]
            if maybe_substr == substr:
                found_substr = maybe_substr
                break
            else:
                substr_pos -= 1
        if not found_substr:
            continue

        # Find non-overlapping start position in case there are multiple
        # consecutive substrings of the same character, e.g. string="!!!!",
        # substr="!!", pos=1 would find the substring at 1, but it actually
        # starts at 0.
        first_substr_pos = substr_pos
        while first_substr_pos > 0:
            first_substr_pos -= 1
            if string[first_substr_pos : first_substr_pos + len_substr] != found_substr:
                first_substr_pos += 1
                break

        # From the first occurrence of `substr`, move as many `substr` lengths
        # forward as do fully fit into the difference between the position where
        # we found `substr` and its first occurrence.
        diff = pos - first_substr_pos
        len_found_substr = len(found_substr)
        substr_pos = first_substr_pos + (int(diff / len_found_substr) * len_found_substr)
        if string[substr_pos : substr_pos + len_found_substr] == found_substr:
            return True, found_substr, substr_pos, diff % len_found_substr

    return False, '', None, None


def _parse(string, delims=DEFAULT_DELIMS, escapes=DEFAULT_ESCAPES, quotes=DEFAULT_QUOTES):
    """
    Yield `Char` instances

    `delims` is a sequence of strings that act as delimiters between arguments.

    `escapes` is a sequence of strings that remove the special meaning of the
    next character.

    `quotes` is a sequences of strings that (mostly) remove the special meaning
    of the characters between two identical quotes. Special cases are:

        1. Any `escape` in front of a normally closing quote is still a special
           character and removes its quoting power.

        2. As a result, any literal `escape` right before a closing quote must
           be escaped so the closing quote can do its job.
    """
    # log.debug('Parsing %r', string)
    state = {'escape': '', 'quote': '', 'closing_quote_end_index': -1}
    for i,char in enumerate(string):
        # log.debug('<%s>', string[:i] + '[' + string[i] + ']' + string[i+1:])
        is_delim, delim, delim_index, delim_offset = _on_any_substr(string, i, delims)
        # if is_delim: log.debug('    Delimiter at %r: %r', delim_index, delim)
        is_quote, quote, quote_index, quote_offset = _on_any_substr(string, i, quotes)
        # if is_quote: log.debug('    Quote at %r: %r', quote_index, quote)
        is_escape, escape, escape_index, escape_offset = _on_any_substr(string, i, escapes)
        # if is_escape: log.debug('    Escape at %r: %r', escape_index, escape)

        if not state['escape']:
            next_is_quote, next_quote, next_quote_index, next_quote_offset = _on_any_substr(string, i + 1, quotes)
            next_is_escape, next_escape, next_escape_index, next_escape_offset = _on_any_substr(string, i + 1, escapes)

            # Backslash is plain text between quotes except when it escapes the
            # enclosing quote (e.g. "foo \" bar" -> [foo " bar])
            if is_escape and (not state['quote'] or next_quote == state['quote'] or next_is_escape):
                yield Char(char, string=escape,
                           is_special=True, is_escaped=state['escape'], is_quoted=state['quote'],
                           escape=escape, quote=state['quote'])
                # Enable escaping state if this is the last character of the
                # escape string (escape characters are not marked as escaped)
                if i >= escape_index + len(escape) - 1:
                    # Only mark escape characters that need it
                    next_is_delim, _, _, _ = _on_any_substr(string, i + 1, delims)
                    if next_is_delim or next_is_quote or next_is_escape:
                        # log.debug('    Enabling escape state')
                        state['escape'] = escape
                #     else:
                #         log.debug('    No need to escape escape next character')
                # else:
                #     log.debug('    Waiting for escape string to end')
                continue

            elif is_quote:
                if state['quote'] == quote:
                    # Closing quote
                    # log.debug('    Found closing quote: %r', quote)
                    state['quote'] = ''
                    yield Char(char, string=quote,
                               is_special=True, is_escaped=state['escape'], is_quoted=state['quote'],
                               quote=quote, escape=state['escape'])
                    # The next character might still be part of the closing quote
                    state['closing_quote_end_index'] = i + len(quote) - 1
                    continue

                elif not state['quote']:
                    # Opening quote
                    yield Char(char, string=quote,
                               is_special=True, is_escaped=state['escape'], is_quoted=state['quote'],
                               quote=quote, escape=state['escape'])

                    # Enable quoting state if this is the last character of the
                    # quote string (quote characters are not marked as quoted)
                    # and if we are not on a closing quote
                    if i >= quote_index + len(quote) - 1 and i > state['closing_quote_end_index']:
                        # log.debug('    Found opening quote: %r', quote)
                        state['quote'] = quote
                    continue

        if is_delim:
            # log.debug('    %r is part of delimiter: %r', char, delim)
            special_string = delim
            special_string_offset = delim_offset
        elif is_escape:
            # log.debug('    %r is part of escape: %r', char, escape)
            special_string = escape
            special_string_offset = escape_offset
        elif is_quote:
            # log.debug('    %r is part of quote: %r', char, quote)
            special_string = quote
            special_string_offset = quote_offset
        else:
            # log.debug('    %r is not part of a special string', char)
            special_string = ''
            special_string_offset = 0

        is_special = is_delim and not state['escape'] and not state['quote']
        yield Char(char, string=special_string,
                   is_special=is_special, is_escaped=state['escape'], is_quoted=state['quote'],
                   delim=delim, escape=state['escape'] or escape, quote=state['quote'] or quote)

        if state['escape'] and special_string and special_string_offset >= len(special_string) - 1:
            # After the first character of a multichar special string is
            # processed, we must keep the escape state alive until all
            # characters of that special string are processed.
            # log.debug('    Disabling escape state')
            state['escape'] = ''


def escape(string, curpos=None, delims=DEFAULT_DELIMS, escapes=DEFAULT_ESCAPES, quotes=DEFAULT_QUOTES):
    """
    Prepend a backslash to special characters: backslash, space, single quote,
    double quote

    If `curpos` is not None, it is the position of the cursor in `string`, and
    the return value is a 2-tuple of (<escaped string>, <new cursor position>).
    """
    string_escaped = []
    curpos_diff = 0
    for i,char in enumerate(_parse(string, delims=delims, escapes=escapes, quotes=quotes)):
        # Escape special or escaped characters
        # The fact that a character is escaped means it is special by nature
        # (non-special characters never have `is_escaped` set to True)
        if char.is_special or char.is_escaped:
            string_escaped.append(escapes[0])
            if curpos is not None and i < curpos:
                curpos_diff += 1
        string_escaped.append(char)

    if curpos is None:
        return ''.join(string_escaped)
    else:
        return ''.join(string_escaped), curpos + curpos_diff


def quote(string, curpos=None, delims=DEFAULT_DELIMS, escapes=DEFAULT_ESCAPES, quotes=DEFAULT_QUOTES):
    """
    Quote `string` if it contains special characters: backslash, space, single
    quote, double quote

    If `curpos` is not None, it is the position of the cursor in `string`, and
    the return value is a 2-tuple of (<quoted string>, <new cursor position>).
    """
    def quote_string():
        # No special strings means no quoting needed
        if all(special_string not in string
               for special_strings in (delims, escapes, quotes)
               for special_string in special_strings):
            return string, '', ''

        # Use the first quote that doesn't exist in `string`
        for quote in quotes:
            if quote not in string:
                return ''.join((quote, string, quote)), quote, ''

        # All quotes exist somewhere in `string` - use the first quoting style
        # and escape all its occurrences
        quote = quotes[0]
        return ''.join((quote,
                        string.replace(quote, escapes[0] + quote),
                        quote)), quote, escapes[0]

    new_string, quote, escape = quote_string()
    # log.debug('Quoted <%s> with <%s>: <%s>', string, quote, new_string)

    if curpos is None:
        return new_string
    else:
        if len(new_string) == len(string):
            # No quotes were added
            return new_string, curpos
        elif curpos <= 0:
            # If cursor was on the first character, put it after the opening quote
            return new_string, len(quote)
        elif curpos >= len(string):
            # If cursor was right of the last character, put it after the closing quote
            return new_string, len(new_string)
        else:
            # Parse `string` up to its cursor position.  Every `quote` we find
            # must be escaped, so we can just add `escape`'s length for each
            # `quote` we find.
            # log.debug('Original string up to %r: %r', curpos, string[:curpos])
            new_curpos = curpos + len(quote)  # Move cursor inside the quotes
            len_escape = len(escape)
            for char in _parse(string[:curpos], delims=delims, escapes=escapes, quotes=quotes):
                if char.string == quote:
                    # log.debug('  Adding len(%r) to cursor position %r: %r', escape, new_curpos, new_curpos+len_escape)
                    new_curpos += len_escape
            # log.debug('New cursor position: %r', new_string[:new_curpos] + '|' + new_string[new_curpos:])
            return new_string, new_curpos


def is_escaped(string, delims=DEFAULT_DELIMS, escapes=DEFAULT_ESCAPES, quotes=DEFAULT_QUOTES):
    """
    Return whether `string` uses backslashes to escape special characters
    """
    for char in _parse(string, delims=delims, escapes=escapes, quotes=quotes):
        if char.is_special:
            if char.string in escapes: return True
            if char.string in quotes: return False
    return False


def plaintext(string, curpos=None, delims=DEFAULT_DELIMS, escapes=DEFAULT_ESCAPES, quotes=DEFAULT_QUOTES):
    """
    Remove any quotes or backslashes from `string` unless they are escaped or
    quoted

    If `curpos` is not None, it is the position of the cursor in `string`, and
    the return value is a 2-tuple of (<plaintext string>, <new cursor position>).
    """
    string_lstripped = string.lstrip()
    if curpos is not None:
        # Adjust cursor position to account for removed leading whitespace
        curpos = max(0, curpos - (len(string) - len(string_lstripped)))

    # Find index of first character of trailing whitespace
    trailing_spaces = len(string_lstripped) - len(string_lstripped.rstrip())
    trailing_spaces_pos = len(string_lstripped) - (trailing_spaces)

    literal = []
    new_curpos = curpos
    for i,char in enumerate(_parse(string_lstripped, delims=delims, escapes=escapes, quotes=quotes)):
        if not char.is_special:
            literal.append(char)
        elif char.is_delim and i < trailing_spaces_pos:
            # Only append non-trailing spaces, i.e. spaces between
            # characters. (Leading spaces have already been stripped.)
            literal.append(char)
        elif char.is_escaped:
            literal.append(char)
        elif new_curpos is not None and new_curpos > 0 and i < curpos:
            new_curpos -= 1

    if new_curpos is None:
        return ''.join(literal)
    else:
        return ''.join(literal), min(new_curpos, len(literal))


def tokenize(cmdline, maxdelims=None, delims=DEFAULT_DELIMS, escapes=DEFAULT_ESCAPES, quotes=DEFAULT_QUOTES):
    """
    Split `cmdline` into list of tokens

    All special characters are retained:

    >>> ''.join(tokenize(cmdline)) == cmdline

    Delimiters are included as separate tokens.
    """
    tokens = []
    token = []
    chars = tuple(_parse(cmdline, delims=delims, escapes=escapes, quotes=quotes))
    maxdelims = float('inf') if maxdelims is None else maxdelims

    def get_token_and_delim_from_end(token_and_delim, delim):
        len_delim = len(delim)
        token = ''.join(token_and_delim[:-len_delim])
        delim = ''.join(token_and_delim[-len_delim:])
        return token, delim

    # log.debug('Tokenizing: %r', cmdline)
    for i,char in enumerate(chars):
        # log.debug('Char: %r', char)

        curdelims = int(len(tokens) / 2)
        if char.is_delim and char.is_special and curdelims < maxdelims:
            # Character is part of a delimiter string that is not escaped or quoted
            prev_char = chars[i - 1] if i > 0 else Char('')

            if not prev_char.is_delim or not prev_char.is_special:
                # This is the first character of a delimiter.  Append the
                # current (non-delimiter) token first before starting a new
                # delimiter token with this character.
                # log.debug('  Delimiter starts at %r: %r: %r', i, char, cmdline[i:])
                if token:
                    tokens.append(''.join(token))
                    # log.debug('    Added %r: %r', tokens[-1], tokens)
                token.clear()
                token[:] = char
            else:
                # This is not the first character of a delimiter, i.e. the
                # current token is a multi-character delimiter.
                # log.debug('  Delimiter already started: %r (prev.is_delim=%r, prev.is_special=%r)',
                #           prev_char.delim, prev_char.is_delim, prev_char.is_special)
                token.append(char)

            if ''.join(token) == char.delim:
                # This is the last character of a delimiter.  Append the current
                # token (a delimiter) before starting a new empty token.
                # log.debug('  Delimiter ends at %r: %r: %r', i, char, cmdline[i:])
                if token:
                    tokens.append(''.join(token))
                    # log.debug('    Added %r: %r', tokens[-1], tokens)
                    token.clear()

            # log.debug('    New token: %r', ''.join(token))
            continue

        token.append(char)
        # log.debug('New token: %r', ''.join(token))

    if token or not tokens:
        tokens.append(''.join(token))
        # log.debug('Added last token: %r', tokens[-1])

    # log.debug('Tokenized: %r', tokens)
    return tokens


def get_position(tokens, curpos):
    """
    Find focused index in `tokens`

    `tokens` is a list of tokens as returned by `tokenize`.

    `curpos` is the cursor position in the string of concatenated `tokens`.

    Return the index of the focused token in `tokens` and the cursor position
    within that token
    """
    # Default cursor position is after the last character of the last argument
    curtok_index = len(tokens) - 1
    curtok_curpos = len(tokens[curtok_index])
    for i,token in enumerate(tokens):
        token_len = len(token)
        curpos -= token_len
        if curpos <= 0:
            curtok_index = i
            curtok_curpos = token_len + curpos
            break
    return curtok_index, curtok_curpos


def remove_delims(tokens, curtok_index, curtok_curpos, delims=DEFAULT_DELIMS):
    """
    Remove any delimiters from `tokens` and adjust `curtok_index` and
    `curtok_curpos` accordingly

    `curtok_index` is adjusted so that the cursor stays on the same argument.

    `curtok_curpos` is adjusted only if the cursor is on a delimiter, in which
    case it is moved to the next token if it's behind the last character of the
    delimiter and the next token is not a delimiter or the previous token
    otherwise.

    Return new `tokens`, `curtok_index` and `curtok_curpos`
    """
    # log.debug('  Input: %r, %r, %r', tokens, curtok_index, curtok_curpos)
    new_tokens = []
    new_curtok_index = curtok_index
    new_curtok_curpos = curtok_curpos
    for i,token in enumerate(tokens):
        if token not in delims:
            # log.debug('    Token %r: %r: Not a delimiter', i, token)
            new_tokens.append(token)

        elif i < curtok_index:
            # If token is a delimiter, reduce `new_curtok_index` as long as we
            # haven't passed the `curtok_index` yet
            new_curtok_index -= 1
            # log.debug('    Token %r: %r: Reduced new_curtok_index by 1: %r', i, token, new_curtok_index)

        elif i == curtok_index:
            # Cursor is on a delimiter
            next_token = tokens[i + 1] if i + 1 < len(tokens) else None
            cursor_after_last_char = curtok_curpos >= len(tokens[curtok_index])
            if next_token is not None and next_token not in delims and cursor_after_last_char:
                # Move to to first character of next token
                new_curtok_curpos = 0
                # log.debug('    Token %r: %r: Moved cursor to next token: %r', i, token, next_token)
            else:
                # Move to to last character of previous token
                if new_tokens:
                    new_curtok_index = max(0, new_curtok_index - 1)
                    new_curtok_curpos = len(new_tokens[new_curtok_index])
                    # log.debug('    Token %r: %r: Moved cursor to previous token: %r', i, token, new_tokens[new_curtok_index])
                else:
                    new_curtok_index = new_curtok_curpos = 0

    # log.debug('  Output: %r, %r, %r', new_tokens, new_curtok_index, new_curtok_curpos)
    return new_tokens, new_curtok_index, new_curtok_curpos


def avoid_delims(tokens, curtok_index, curtok_curpos, delims=DEFAULT_DELIMS):
    """
    If the cursor is on the first character of a delimiter, move it behind the
    last character of any previous non-delimiter token.  If the cursor is on any
    other character of a delimiter, move it on the first character of the next
    non-delimiter token.

    If no previous non-delimiter token exists, move on the first character of
    the next non-delimiter token.  And vice-versa, if no next non-delimiter
    token exists, move behind the last character of the previous non-delimiter
    token. If `tokens` contains only delimiters, keep the cursor position as it
    is.

    Return new `tokens`, `curtok_index` and `curtok_curpos`
    """
    curtok = tokens[curtok_index]
    if curtok not in delims:
        return tokens, curtok_index, curtok_curpos
    else:
        class ImpossibleMove(Exception): pass

        def forward():
            # log.debug('      Finding first non-delimiter token in %r', tokens[curtok_index+1:])
            for i,tok in enumerate(tokens[curtok_index + 1:], start=1):
                if tok not in delims:
                    # log.debug('      Avoiding delimiter - moving %d tokens forward', i)
                    return curtok_index + i, 0
            raise ImpossibleMove()

        def backward():
            # log.debug('      Finding last non-delimiter token in %r', tokens[:curtok_index])
            for i,tok in enumerate(reversed(tokens[:curtok_index]), start=1):
                if tok not in delims:
                    # log.debug('      Avoiding delimiter - moving %d tokens backward', i)
                    return curtok_index - i, len(tokens[curtok_index - i])
            raise ImpossibleMove()

        def move(*movements):
            for func in movements:
                # log.debug('    Trying to move %s', func.__name__)
                try:
                    return func()
                except ImpossibleMove:
                    # log.debug('      No dice')
                    pass
            return curtok_index, curtok_curpos

        prev_token = tokens[curtok_index - 1] if curtok_index > 0 else None
        # log.debug('  Input: %r, %r, %r', tokens, curtok_index, curtok_curpos)

        # Move to previous token if we're on the first character of a delimiter
        # and the previous token is a non-delimiter
        # ['ls', '| ', 'foo'] -> ['ls|', ' ', 'foo']
        if curtok_curpos <= 0 and prev_token is not None and prev_token not in delims:
            curtok_index, curtok_curpos = move(backward, forward)

        # Move to next token if we're on the last character of a delimiter
        # ['ls', ' |', 'foo'] -> ['ls', ' ', '|foo']
        else:
            curtok_index, curtok_curpos = move(forward, backward)

    # log.debug('  Output: %r, %r, %r', tokens, curtok_index, curtok_curpos)
    return tokens, curtok_index, curtok_curpos


def maybe_insert_empty_token(tokens, curtok_index, curtok_curpos, delims=DEFAULT_DELIMS):
    """
    If the cursor is on a delimiter at the beginning or end of the command line,
    an empty token is inserted there.

    If the cursor is on the first character of a delimiter and the previous
    token is also a delimiter, an empty token is inserted at `curtok_index`.

    If the cursor is on the last character of a delimiter and the next token is
    also a delimiter, an empty token is inserted at `curtok_index+1`.

    If the cursor is on any but the first or last character of a multi-character
    delimiter, the delimiter is split at `curtok_curpos` and an empty token is
    inserted there.

    If an empty token is inserted, the cursor is always placed on it.

    Return new `tokens`, `curtok_index` and `curtok_curpos`
    """
    # log.debug('  Input: %r, %r, %r', tokens, curtok_index, curtok_curpos)
    curtok = tokens[curtok_index]
    curtok_index_max = len(tokens) - 1
    curtok_curpos_max = len(curtok)

    if curtok in delims:
        on_first_char = curtok_curpos == 0
        on_last_char = curtok_curpos == curtok_curpos_max
        prevtok = tokens[curtok_index - 1] if curtok_index >= 1 else None
        nexttok = tokens[curtok_index + 1] if curtok_index < curtok_index_max else None

        if curtok_index == 0:
            # log.debug('    Inserting empty token before the first')
            tokens = ['', *tokens]
            curtok_curpos = 0

        elif curtok_index == curtok_index_max:
            # log.debug('    Inserting empty token after the last')
            tokens = [*tokens, '']
            curtok_index = curtok_index_max + 1
            curtok_curpos = 0

        elif on_first_char and prevtok in delims:
            # ['foo', '/', '|/', 'bar'] -> ['foo', '/', '|', '/', 'bar']
            # log.debug('    Inserting empty token before the current')
            tokens = [*tokens[:curtok_index], '', *tokens[curtok_index:]]
            curtok_curpos = 0

        elif on_last_char and nexttok in delims:
            # ['foo', '/|', '/', 'bar'] -> ['foo', '/', '|', '/', 'bar']
            # log.debug('    Inserting empty token after the current')
            curtok_index += 1
            tokens = [*tokens[:curtok_index], '', *tokens[curtok_index:]]
            curtok_curpos = 0

        elif not on_first_char and not on_last_char:
            # ['foo', '!|=', 'bar'] -> ['foo', '!', '|', '=', 'bar']
            # log.debug('    Inserting empty token in multi-char delimiter')
            subtoks = (curtok[:curtok_curpos], '', curtok[curtok_curpos:])
            # log.debug('    Replacing %r with %r', curtok, subtoks)
            tokens = [*tokens[:curtok_index], *subtoks, *tokens[curtok_index + 1:]]
            curtok_index += 1
            curtok_curpos = 0
    #     else:
    #         log.debug('    No need for empty token after %r: %r', curtok, tokens)
    # else:
    #     log.debug('    Not a delimiter: %r', curtok)

    # log.debug('  Output: %r, %r, %r', tokens, curtok_index, curtok_curpos)
    return tokens, curtok_index, curtok_curpos


def _get_paramspec(options, arg):
    for key,spec in options.items():
        if arg in key:
            return spec


def remove_options(args, curarg_index, curarg_curpos, options={}):
    """
    Remove options and parameters from `args` and adjust `curarg_index` and
    `curarg_curpos` if necessary

    `options` is a mapping that maps tuples of option names (e.g. `('--output',
    '-o')`) to one of these values:

        <int> - Remove <int> arguments after these options
        "*"   - Remove all arguments after these options until an argument
                that starts with "-" is encountered

    (Parameters are defined as arguments to an option, e.g. "--option parameter".)

    Return new arguments, index of current argument and cursor position in
    current argument
    """
    # Reduce index of current argument if we remove options to the left of it
    new_curarg_index = curarg_index
    new_curarg_curpos = curarg_curpos
    cur_spec, params_found = None, 0
    args_wo_opts = []

    for i,arg in enumerate(args):
        append_arg = True
        is_option = arg.startswith('-')
        if is_option:
            append_arg = False
            cur_spec, params_found = _get_paramspec(options, arg), 0
        elif cur_spec is not None:
            # Find out whether this argument is a parameter for an option we
            # encountered earlier
            params_found += 1
            if isinstance(cur_spec, int):
                if params_found <= cur_spec:
                    append_arg = False
            elif cur_spec == '*':
                if not is_option:
                    append_arg = False
            else:
                raise ValueError('Invalid parameter spec: %r' % (cur_spec,))

        if append_arg:
            args_wo_opts.append(arg)
        elif curarg_index is not None and curarg_curpos is not None:
            if i < curarg_index:
                # Removing an argument before the current one
                new_curarg_index -= 1
            elif i == curarg_index:
                # Removing the current argument
                new_curarg_curpos = 0

    # Ensure at least one empty argument
    if not args_wo_opts:
        args_wo_opts = ['']
        new_curarg_index = new_curarg_curpos = 0

    # Ensure index and curpos are not out of bounds
    elif curarg_index is not None and curarg_curpos is not None:
        new_curarg_index_max = len(args_wo_opts) - 1
        new_curarg_curpos_max = len(args_wo_opts[-1])
        if new_curarg_index > new_curarg_index_max:
            new_curarg_index = new_curarg_index_max
            new_curarg_curpos = new_curarg_curpos_max
        elif new_curarg_curpos > new_curarg_curpos_max:
            new_curarg_curpos = new_curarg_curpos_max

    return args_wo_opts, new_curarg_index, new_curarg_curpos


def get_nth_posarg_index(n, args, options={}):
    """
    Return index in `args` of the `n`th positional argument.  Positional
    arguments don't start with "-" and are not parameters for options.

    Return None if there are no positional arguments in `args`.

    See `remove_options` for documentation about `options`.
    """
    posargs_found = -1
    cur_spec, params_found = None, 0
    for i,arg in enumerate(args):
        is_option = arg.startswith('-')
        is_posarg = False

        if is_option:
            cur_spec, params_found = _get_paramspec(options, arg), 0
        elif cur_spec is not None:
            # Find out whether this argument is a parameter for an option
            params_found += 1

            if isinstance(cur_spec, int):
                if params_found > cur_spec:
                    is_posarg = True
                    posargs_found += 1
            elif cur_spec == '*':
                if not is_option:
                    is_posarg = True
                    posargs_found += 1
            else:
                raise ValueError('Invalid parameter spec: %r' % (cur_spec,))

        else:
            # Option is not specified in `options` but it's still an option
            is_posarg = True
            posargs_found += 1

        if is_posarg and posargs_found == n - 1:
            return i
    return None


def get_current_cmd(tokens, curtok_index, ops):
    """
    Extract tokens before and after the currently focused token up to any
    operators or the start/end

    Return slice of those tokens and the index of the currently focused token in
    that slice
    """
    def is_op(token):
        return any(token == op for op in ops)

    if is_op(tokens[curtok_index]):
        # Cursor is on operator, so no command to return
        return (None, None)
    else:
        first_tok = 0
        last_tok = len(tokens) - 1

        # Find operator before focused token
        for i in range(curtok_index - 1, 0, -1):
            token = tokens[i]
            if is_op(token):
                first_tok = i + 1
                break

        # Find operator after focused token
        for i,token in enumerate(tokens[curtok_index:-1]):
            if is_op(token):
                last_tok = curtok_index + i - 1
                break

    sub_curtok_index = curtok_index - first_tok
    return (tokens[first_tok : last_tok + 1], sub_curtok_index)


class Arg(str):
    """Single argument"""

    def __new__(cls, arg, curpos=None):
        obj = super().__new__(cls, arg)
        obj._curpos = curpos
        return obj

    @property
    def before_cursor(self):
        """Return the part before the cursor or the whole argument if `curpos` is None"""
        if self.curpos is not None:
            return Arg(self[:self.curpos], curpos=self.curpos)
        else:
            return self

    @property
    def curpos(self):
        """Cursor position in argument"""
        return self._curpos

    @curpos.setter
    def curpos(self, curpos):
        self._curpos = int(curpos)

    def separate(self, seps, maxseps=None, include_seps=True):
        """
        Split argument at any occurrence of any string in `seps`

        maxseps: Maximum number of separations or None
        include_seps: Whether to include separators in the new list

        Return an `Args` object
        """
        seps = tuple(seps)
        # log.debug('Splitting %r at separators: %r', str(self), seps)
        parts = tokenize(self, delims=seps, maxdelims=maxseps)

        if self.curpos is not None:
            curpart_index, curpart_curpos = get_position(parts, self.curpos)
        else:
            curpart_index, curpart_curpos = 0, 0

        parts, curpart_index, curpart_curpos = maybe_insert_empty_token(parts, curpart_index, curpart_curpos, seps)
        # log.debug('Maybe inserted empty token: %r, %r, %r', parts, curpart_index, curpart_curpos)

        if include_seps:
            if self.curpos is not None:
                parts, curpart_index, curpart_curpos = avoid_delims(parts, curpart_index, curpart_curpos, seps)
                # log.debug('Moved away from separators: %r, %r, %r', parts, curpart_index, curpart_curpos)
        else:
            parts, curpart_index, curpart_curpos = remove_delims(parts, curpart_index, curpart_curpos, seps)
            # log.debug('Removed separators: %r, %r, %r', parts, curpart_index, curpart_curpos)

        if self.curpos is not None:
            return Args(parts, curpart_index, curpart_curpos)
        else:
            return Args(parts)

    def __getitem__(self, item):
        if isinstance(item, slice):
            subarg = super().__getitem__(item)
            if item.step is not None:
                raise RuntimeError('Slicing with steps is not implemented yet')
            elif self.curpos is None:
                return Arg(subarg)
            elif ((item.start is None or item.start <= self.curpos) and
                  (item.stop is None or item.stop > self.curpos)):
                curpos = self.curpos - (item.start or 0)
            else:
                curpos = None
            return Arg(subarg, curpos=curpos)
        else:
            curpos = 0 if item == self.curpos else None
            return Arg(super().__getitem__(item), curpos=curpos)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return super().__eq__(other) and self.curpos == other.curpos

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __repr__(self):
        string = '%s(%r' % (type(self).__name__, str(self),)
        if self.curpos is not None:
            string += ', curpos=%d' % (self.curpos,)
        return string + ')'


class Args(tuple):
    """
    Tuple of arguments (tokens without delimiters) with cursor position and
    some convenience properties
    """
    @classmethod
    def from_tokens(cls, tokens, curtok_index, curtok_curpos, delims=DEFAULT_DELIMS):
        # log.debug('Creating Args from tokens: %r, %r, %r', tokens, curtok_index, curtok_curpos)

        # Ensure at least one empty string
        if not tokens:
            tokens = ('',)

        tokens, curtok_index, curtok_curpos = maybe_insert_empty_token(tokens, curtok_index, curtok_curpos, delims)
        # log.debug('Maybe inserted empty token: %r, %r, %r', tokens, curtok_index, curtok_curpos)

        tokens, curtok_index, curtok_curpos = remove_delims(tokens, curtok_index, curtok_curpos, delims)
        # log.debug('Removed separators: %r, %r, %r', tokens, curtok_index, curtok_curpos)

        # Convert everything to plaintext (e.g. remove backslashes from escaped characters)
        for i in range(len(tokens)):
            if i == curtok_index:
                tokens[i], curtok_curpos = plaintext(tokens[i], curtok_curpos)
            else:
                tokens[i] = plaintext(tokens[i])

        return cls(tokens, curtok_index, curtok_curpos)

    def __new__(cls, args, curarg_index=None, curarg_curpos=None):
        def gen():
            for i,arg in enumerate(args):
                if i == curarg_index:
                    yield Arg(arg, curpos=curarg_curpos)
                else:
                    yield Arg(arg)
        obj = super().__new__(cls, gen())
        obj._curarg_index = curarg_index
        if curarg_index is not None and obj.curarg_curpos is None:
            obj.curarg_curpos = 0
        return obj

    @property
    def curarg(self):
        """Currently focused argument"""
        curarg_index = self._curarg_index
        if curarg_index is not None:
            return self[curarg_index]

    @property
    def before_curarg(self):
        """All arguments up to current argument"""
        if self._curarg_index is None:
            return self
        else:
            return self[:self._curarg_index]

    @property
    def curarg_index(self):
        """Index of currently focused argument"""
        return self._curarg_index

    @curarg_index.setter
    def curarg_index(self, index):
        self._curarg_index = index

    @property
    def curarg_curpos(self):
        """Cursor position in currently focused argument"""
        curarg = self.curarg
        if curarg is not None:
            return curarg.curpos

    @curarg_curpos.setter
    def curarg_curpos(self, curpos):
        curarg = self.curarg
        if curarg is not None:
            curarg.curpos = curpos

    def posargs(self, options={}):
        """
        Return copy without options and parameters for options

        `options` is a mapping that maps tuples of option names
        (e.g. `('--output', '-o')`) to one of these values:

            <int> - Remove <int> arguments after these options
            "*"   - Remove all arguments after these options until an argument
                    that starts with "-" is encountered
        """
        return Args(*remove_options(self, self.curarg_index, self.curarg_curpos, options))

    def params(self, option, maxparams=None):
        """
        Return list of parameters for `option`

        `option` is either the name of an option (e.g. "-h" or "--help") or an
        iterable of names.

        `max` is the maximum number of parameters to return.
        """
        # Find index of option
        option_index = None
        if isinstance(option, str):
            for i,arg in enumerate(self):
                if arg == option:
                    option_index = i
                    break
        elif isinstance(option, abc.Iterable):
            opts = tuple(option)
            for i,arg in enumerate(self):
                if any(arg == o for o in opts):
                    option_index = i
                    break
        # Get parameters that follow option
        params = []
        curarg_index = None
        curarg_curpos = None
        if option_index is not None:
            stop = option_index + maxparams + 1 if maxparams is not None else len(self)
            for i in range(option_index + 1, stop):
                try:
                    arg = self[i]
                except IndexError:
                    break
                if arg and arg[0] == '-':
                    break
                params.append(arg)
                if arg.curpos is not None:
                    curarg_curpos = arg.curpos
                    curarg_index = i - option_index - 1
        return Args(params, curarg_index, curarg_curpos)

    def nth_posarg_index(self, n, options={}):
        """
        Return index of `n`th positional argument (any argument that isn't an option
        or a parameter for an option)

        Return None if there are no positional arguments.

        See `posargs` for documentation on the `options` argument.
        """
        return get_nth_posarg_index(n, self, options)

    def remove_empty(self):
        """Return new Args object with any empty ("") arguments removed"""
        new_args = []
        new_curarg_index = self.curarg_index
        new_curarg_curpos = self.curarg_curpos
        for i,arg in enumerate(self):
            if len(arg) >= 1:
                new_args.append(arg)
            elif self.curarg_index > i:
                new_curarg_index -= 1
        return Args(new_args,
                    curarg_index=min(new_curarg_index, len(new_args) - 1),
                    curarg_curpos=new_curarg_curpos)

    def __getitem__(self, item):
        if isinstance(item, slice):
            subargs = super().__getitem__(item)
            if not subargs:
                # Args are never really empty
                subargs = ('',)
                curarg_index = curarg_curpos = None
            elif item.step is not None:
                raise RuntimeError('Slicing with steps is not implemented yet')
            elif self.curarg_index is None or self.curarg_curpos is None:
                # No cursor position specified
                curarg_index = curarg_curpos = None
            else:
                start = item.start if item.start is not None else 0
                stop  = item.stop  if item.stop  is not None else len(self)
                if ((item.start is None or item.start <= self.curarg_index) and
                    (item.stop is None or item.stop > self.curarg_index)):
                    # Cursor is between `start` and `stop` or on `start`
                    curarg_index = self.curarg_index - start
                    curarg_curpos = self.curarg_curpos
                elif self.curarg_index == start - 1 and self.curarg_curpos == len(self.curarg):
                    # Cursor is after the last character of the argument before
                    # `start`, i.e. just before our slice.  This is essentially
                    # the same position as on the first character of the
                    # argument at `start`.
                    # Example: ['a|', 'b', 'c'][1:]  ->  ['|b', 'c']
                    curarg_index = self.curarg_index - start + 1
                    curarg_curpos = 0
                elif self.curarg_index == stop and self.curarg_curpos == 0:
                    # Cursor is on the first character of the argument at
                    # `stop`, i.e. just after our slice.  This is essentially
                    # the same position as after the last character of the
                    # argument at `stop`.
                    # Example: ['a', 'b', '|c'][:2]  ->  ['a', 'b|']
                    curarg_index = self.curarg_index - start - 1
                    curarg_curpos = len(subargs[-1])
                    curarg_curpos = len(subargs[-1])
                else:
                    curarg_index = curarg_curpos = None
            return Args(subargs, curarg_index=curarg_index, curarg_curpos=curarg_curpos)
        else:
            return super().__getitem__(item)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (super().__eq__(other) and
                self.curarg_index == other.curarg_index and
                self.curarg_curpos == other.curarg_curpos)

    def __hash__(self):
        return super().__hash__()

    def __repr__(self):
        string = '%s(%s' % (type(self).__name__, tuple(str(arg) for arg in self))
        if self.curarg_index is not None:
            string += ', curarg_index=%d' % (self.curarg_index,)
        if self.curarg_curpos is not None:
            string += ', curarg_curpos=%d' % (self.curarg_curpos,)
        return string + ')'



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt


def make_X_FILTER_spec(filtername, or_focused=False, more_text='', **kwargs):
    spec = {'names': (filtername.upper() + ' FILTER',),
            'description': ('%s filter expression (see %s FILTERS section in `help filters`)' %
                            (filtername.capitalize(), filtername.upper()))}
    if or_focused:
        spec['description'] += ' or focused %s in the TUI if omitted' % filtername.lower()
    if more_text:
        spec['description'] += '; %s' % more_text
    spec.update(**kwargs)
    return spec


def make_SCRIPTING_doc(cmdname):
    return (("If invoked as a command line argument and the output does not "
             "go to a TTY (i.e. the terminal size can't be determined), "
             "the output is optimized for scripting.  Numbers are "
             "unformatted, columns are separated by a horizontal tab "
             "character ('\\t') and headers are not printed."),
            "",
            ("To enforce human-readable, formatted output, set the environment "
             "variables COLUMNS and LINES."),
            "",
            "\t$ \tCOLUMNS=80 LINES=24 {{__appname__}} {CMDNAME} | less -R".format(CMDNAME=cmdname))


def make_SORT_ORDERS_doc(sortercls, option, setting, append=()):
    doc = [('The following sort orders can be specified with the {option} option '
            'or the "{setting}" setting:').format(option=option, setting=setting),
           '']

    for sname,s in sorted(sortercls.SORTSPECS.items()):
        snames = ', '.join((sname,) + s.aliases)
        doc.append('\t{}\t - \t{}'.format(snames, s.description))

    doc.extend(('',
                'Multiple sort orders are separated with "," without spaces.',
                '',
                'Sorting is reversed if the sort order is prepended by "!" or ".".',
                '',
                ('If "%s" is not given explicitly, it is always prepended to '
                 'the list of sort orders.') % sortercls.DEFAULT_SORT))
    if append:
        doc.extend(('',) + append)
    return tuple(doc)


def make_COLUMNS_doc(columnspecs, option, setting, append=()):
    return (('The following columns can be specified with the {option} option '
             'or the "{setting}" setting:').format(option=option, setting=setting),
            '',
            '\t%s' % ', '.join(sorted(columnspecs)),
            '',
            'Columns are separated with "," without spaces.') + append



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import collections

import urwid

from ..main import redraw_screen
from urwid import ScrollBar
from ..table import ColumnHeaderWidget, Table
from ..tuiobjects import bottombar


class Style():
    """Map standard attributes to those defined in a urwid palette

    prefix: common prefix of all attributes
    extras: additional attributes (e.g. 'header')
    modes: additional subsections with 'focused' and 'unfocused' attributes
           (e.g. 'highlighted')
    focusable: True if 'focused' attributes should be mapped, False otherwise
    """

    def __init__(self, prefix, modes=(), extras=(), focusable=True):
        self._attribs = {
            'unfocused': self.dotify(prefix, 'unfocused')
        }
        for extra in extras:
            self._attribs[extra] = self.dotify(prefix, extra)

        if focusable:
            self._attribs['focused'] = self.dotify(prefix, 'focused')

        for mode in modes:
            self._attribs[mode + '.focused'] = self.dotify(prefix, mode, 'focused')
            self._attribs[mode + '.unfocused'] = self.dotify(prefix, mode, 'unfocused')

    def attrs(self, mode=None, focused=False):
        """Get attributes as specified in the urwid palette

        mode: one of the modes specified during initialization
        focused: If True the '...focused' attributes are returned,
                 '...unfocused otherwise
        """
        mode = '' if not mode else mode
        name = self.dotify(mode, 'focused' if focused else 'unfocused')
        if name in self._attribs:
            return self._attribs[name]
        else:
            return self._attribs[mode]

    @property
    def focus_map(self):
        """Map of all '...unfocused' -> '...focused' attributes"""
        focus_map = {}
        attribs = self._attribs
        for name in attribs:
            if name.endswith('unfocused'):
                name_focused = name[:-9] + 'focused'
                focus_map[attribs[name]] = attribs[name_focused]
        return focus_map

    @staticmethod
    def dotify(*strings):
        """Join non-empty strings with a '.'"""
        return '.'.join(x for x in (strings) if x)


class CellWidgetBase(urwid.WidgetWrap):
    """Base class for cells in items in Torrent/File/Peer/... lists"""

    style = collections.defaultdict(lambda: 'default')
    header = urwid.AttrMap(ColumnHeaderWidget(left='', right=''), 'header')
    width = ('weight', 100)
    align = 'right'

    def __init__(self):
        self.value = None
        self.text = urwid.Text('', wrap=self.wrap, align=self.align)
        self.attrmap = urwid.AttrMap(self.text, self.style.attrs('unfocused'))
        return super().__init__(self.attrmap)

    def update(self, data):
        self.data = data
        new_value = self.get_value()
        new_text = str(new_value)
        if self.text.text != new_text:
            self.value = new_value
            self.text.set_text(new_text)
            attr = self.style.attrs(self.get_mode(), focused=False)
            self.attrmap.set_attr_map({None: attr})

    def get_mode(self):
        return None

    @classmethod
    def set_header(cls, left=None, right=None):
        if left is not None:
            cls.header.base_widget.left = str(left)
        if right is not None:
            cls.header.base_widget.right = str(right)


class ItemWidgetBase(urwid.WidgetWrap):
    """Base class for items in Torrent/File/Peer/... lists"""

    # Derived classes must set these class attributes; lists with unfocusable
    # items (e.g. peer lists) don't have to set palette_focused and
    # columns_focus_map.
    columns_focus_map = NotImplemented
    palette_unfocused = NotImplemented
    palette_focused   = NotImplemented

    def __init__(self, data, cells):
        self._data = data    # Info of torrent/tracker/file/peer/... as mapping
        self._cells = cells  # Group instance that combines widgets horizontally

        # Create focusable or unfocusable item widget
        if self.columns_focus_map is not NotImplemented:
            item_widget = urwid.AttrMap(
                urwid.AttrMap(cells, attr_map=None, focus_map=self.columns_focus_map),
                self.palette_unfocused, self.palette_focused
            )
        else:
            item_widget = urwid.AttrMap(cells, self.palette_unfocused)
        urwid.WidgetWrap.__init__(self, item_widget)

        # Initialize cell widgets
        self.update(data)

    def update(self, data):
        for widget in self._cells.widgets:
            if hasattr(widget, 'update'):
                widget.update(data)
        self._data = data

    @property
    def id(self):
        """Unique, hashable ID of the displayed item"""
        raise NotImplementedError()

    @property
    def data(self):
        """Displayed data in dictionary form"""
        return self._data

    @property
    def is_marked(self):
        """Whether this item has been marked by the user"""
        if self._cells.exists('marked'):
            return self._cells.marked.is_marked
        else:
            return False

    @is_marked.setter
    def is_marked(self, is_marked):
        if self._cells.exists('marked'):
            self._cells.marked.is_marked = bool(is_marked)


class ListWidgetBase(urwid.WidgetWrap):
    """Base class for Torrent/File/Peer/... lists"""

    # Derived classes must set these class attributes
    tuicolumns      = NotImplemented
    ListItemClass   = NotImplemented
    keymap_context  = NotImplemented
    palette_name    = NotImplemented
    focusable_items = False

    def __init__(self, srvapi, keymap, columns=None, sort=None, title=None):
        self._srvapi = srvapi
        self._keymap = keymap

        if self.focusable_items:
            self._ListItemClass = keymap.wrap(self.ListItemClass, context=self.keymap_context)
        else:
            self._ListItemClass = self.ListItemClass

        self._data_dict = None
        self._marked = set()

        self._existing_widgets = set()
        self._hidden_widgets = set()

        self._sort = sort
        self._sort_orig = sort

        self._title_name = title
        self.title_updater = None

        self._table = Table(**self.tuicolumns)
        self._table.columns = columns or ()

        if self.focusable_items:
            walker = urwid.SimpleFocusListWalker([])
        else:
            walker = urwid.SimpleListWalker([])
        self._listbox = keymap.wrap(urwid.ListBox, context=self.keymap_context + 'list')(walker)

        listbox_sb = urwid.AttrMap(
            ScrollBar(urwid.AttrMap(self._listbox, self.palette_name)),
            self.palette_name + '.scrollbar'
        )
        pile = urwid.Pile([
            ('pack', urwid.AttrMap(self._table.headers, self.palette_name + '.header')),
            listbox_sb
        ])
        super().__init__(pile)

    def __repr__(self):
        return '<%s %s, #%s>' % (type(self).__name__, self.title, id(self))

    @redraw_screen
    def _invalidate(self):
        if self.title_updater is not None:
            # First argument can be cropped if too long, second argument is fixed
            self.title_updater(self.title, ' [%d]' % self.count)
        super()._invalidate()

    def render(self, size, focus=False):
        # Remember focused item widget in case items get added or removed
        focusedw = self.focused_widget

        if self._data_dict is not None:
            self._update_existing_widgets(self._data_dict)
            self._data_dict = None

        self._hide_or_unhide_widgets()
        self._sort_widgets()

        # Ensure focus doesn't change when items get added or removed
        if focusedw is not None and self.focused_widget is not None and \
           focusedw.id != self.focused_widget.id:
            focused_id = focusedw.id
            for i,w in enumerate(self._listbox.body):
                if w.id == focused_id:
                    self._listbox.focus_position = i
                    break

        # Update number of marked items in this list
        bottombar.marked.update(len(self._marked))

        # focus=True because we always want to highlight the focused item, for
        # example when the CLI is open
        return super().render(size, focus=True)

    def _update_existing_widgets(self, data_dict):
        existing_widgets = self._existing_widgets
        dead_widgets = []

        for w in existing_widgets:  # w = *ItemWidget instance
            id = w.id
            try:
                # Update existing *ItemWidget instances with new data
                w.update(data_dict[id])
                del data_dict[id]
            except KeyError:
                # Item no longer exists in data_dict anymore
                dead_widgets.append(w)

        # Remove dead *ItemWidget instances
        walker = self._listbox.body
        marked = self._marked
        for w in dead_widgets:
            if w in walker:
                walker.remove(w)
            existing_widgets.remove(w)
            marked.discard(w)  # self._marked may have a reference too

        # Any items that haven't been used to update an existing *ItemWidget instance are new
        if data_dict:
            table = self._table
            ListItemClass = self._ListItemClass
            for data_id,data in data_dict.items():
                table.register(data_id)
                row = table.get_row(data_id)
                existing_widgets.add(ListItemClass(data, row))

    def _sort_widgets(self):
        walker = self._listbox.body
        if self._sort is not None:
            try:
                self._sort.apply(walker,
                                 item_getter=lambda w: w.data,
                                 inplace=True)
            except KeyError:
                # This happens when adding a new sort order that needs
                # previously unneeded keys (e.g. "started" needs "time-started",
                # which is normally not used).  The new request is correctly
                # registered in client.trequestpool, but when the async RPC
                # request is made, the asyncio loop yields control to the TUI,
                # which redraws (i.e. sorts) the list with the old widget.data.
                # (I couldn't figure out why this redraw happens.)  Ignoring the
                # KeyError fixes this because as soon as the RPC response gets
                # through, a new redraw is issued and the new sort exists.
                pass

    def _hide_or_unhide_widgets(self):
        walker = self._listbox.body
        existing_widgets = self._existing_widgets
        hidden_ids = tuple(w.id for w in self._limit_items(existing_widgets))
        for w in existing_widgets:
            widget_is_visible = w in walker
            hide_widget = w.id in hidden_ids
            if hide_widget and widget_is_visible:
                walker.remove(w)
                self._hidden_widgets.add(w)
            elif not hide_widget and not widget_is_visible:
                walker.append(w)

        if self.title_updater is not None:
            self.title_updater(self.title, ' [%d]' % self.count)

    def _limit_items(self, existing_widgets):
        """Iterate over filtered widgets"""
        return ()

    def clear(self):
        """Remove all list items"""
        self._table.clear()
        self._listbox.body[:] = ()
        self._listbox._invalidate()
        self._marked.clear()

    def refresh(self):
        """Update list items"""
        raise NotImplementedError


    @property
    def columns(self):
        return self._table.columns

    @columns.setter
    def columns(self, columns):
        self._table.columns = columns

    @property
    def sort(self):
        """*Sorter object or `None` to keep list items unsorted"""
        return self._sort

    @sort.setter
    def sort(self, sort):
        if sort == 'RESET':
            self._sort = self._sort_orig
        else:
            self._sort = sort

    @property
    def count(self):
        """Number of listed items"""
        # If this method was called before rendering, the contents of the
        # listbox widget are inaccurate and we have to use self._data_dict.  But
        # if we're called after rendering, self._data_dict is reset to None and
        # we have to count items in the listbox.
        if self._data_dict is not None:
            return len(self._data_dict)
        else:
            return len(self._listbox.body)

    DEFAULT_TITLE = 'No title'

    @property
    def title_name(self):
        """The base name of the title"""
        return str(self._title_name or self.DEFAULT_TITLE)

    @property
    def title_sort_order(self):
        """The sort order part of the title"""
        if self._sort is not None:
            sortstr = str(self._sort)
            if sortstr is not self._sort.DEFAULT_SORT:
                return '{%s}' % sortstr
        return ''

    @property
    def title(self):
        """Combined `title_name` and `title_sort_order`"""
        parts = [self.title_name]
        sort = self.title_sort_order
        if sort:
            parts.append(sort)
        return ' '.join(parts)


    @property
    def has_marked_column(self):
        return 'marked' in self._table.columns

    def mark(self, toggle=False, all=False):
        """Mark the currently focused item or all items"""
        self._set_mark(True, toggle=toggle, all=all)

    def unmark(self, toggle=False, all=False):
        """Unmark the currently focused item or all items"""
        self._set_mark(False, toggle=toggle, all=all)

    @property
    def marked(self):
        """Generator that yields ItemWidgetBase descendants"""
        secondary_filter = self.secondary_filter
        if secondary_filter is None:
            yield from self._marked
        else:
            for widget in self._marked:
                if secondary_filter.match(widget.data):
                    yield widget

    @property
    def marked_count(self):
        """Number of marked items"""
        return len(self._marked)

    def _set_mark(self, mark, toggle=False, all=False):
        if toggle and self.focused_widget is not None:
            mark = not self.focused_widget.is_marked

        for widget in self._select_items_for_marking(all):
            widget.is_marked = mark
            if mark:
                self._marked.add(widget)
            else:
                self._marked.discard(widget)

    def _select_items_for_marking(self, all):
        if self.focused_widget is not None:
            if all:
                yield from self._listbox.body
            else:
                yield self.focused_widget

    def refresh_marks(self):
        """
        Redraw the "marked" column in all items widgets

        This shouldn't be needed unless the marked character was changed.
        """
        for widget in self._listbox.body:
            widget.is_marked = widget.is_marked


    @property
    def secondary_filter(self):
        """Another filter that can be changed during the lifetime of this list"""
        return None

    @secondary_filter.setter
    def secondary_filter(self, value):
        raise RuntimeError('%r does not have a secondary_filter implementation' % (self,))


    @property
    def items(self):
        """Yield non-hidden widgets from list"""
        for item in self._listbox.body:
            yield item

    @property
    def focused_widget(self):
        """Currently focused widget in list"""
        if self.focusable_items:
            return self._listbox.focus

    @property
    def focused_id(self):
        """ID of the currently focused list item or `None`"""
        focused_widget = self._listbox.focus
        if focused_widget is not None:
            return focused_widget.id

    @property
    def focus_position(self):
        """Focus position (first item is 0; `None` if list is empty)"""
        return self._listbox.focus_position

    @focus_position.setter
    def focus_position(self, focus_position):
        self._listbox.focus_position = min(focus_position, len(self._listbox.body) - 1)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import functools
import itertools

from ..client import filters as filter_clses
from ..client import sorters as sorter_clses
from . import Candidate, Candidates

columns_labels = {'torrents' : 'Torrent List Column',
                  'files'    : 'File List Column',
                  'peers'    : 'Peer List Column',
                  'trackers' : 'Tracker List Column',
                  'settings' : 'Setting List Column'}


sorters_labels = {'TorrentSorter' : 'Torrent Sort Order',
                  'PeerSorter'    : 'Peer Sort Order',
                  'TrackerSorter' : 'Tracker Sort Order',
                  'SettingSorter' : 'Setting Sort Order'}


@functools.lru_cache(maxsize=None)
def get_sorter_cls(clsname):
    return getattr(sorter_clses, clsname)


# All filters use the same operators
filter_compare_ops = filter_clses.TorrentFilter.POSSIBLE_OPERATORS
filter_combine_ops = ('&', '|')
filter_labels = {'TorrentFilter'  : 'Torrent Filter',
                 'FileFilter'    : 'File Filter',
                 'PeerFilter'    : 'Peer Filter',
                 'TrackerFilter' : 'Tracker Filter',
                 'SettingFilter' : 'Setting Filter'}

@functools.lru_cache(maxsize=None)
def filter_names(filter_cls_name):
    """Filter names as Candidate instances with description and aliases"""
    def get_names(filter_cls):
        for name in get_filter_names(filter_cls):
            filter_spec = get_filter_spec(filter_cls, name)
            desc = filter_spec.description
            alias_str = ','.join(filter_spec.aliases)
            yield Candidate(name, in_parens=alias_str, Description=desc)

    filter_cls = get_filter_cls(filter_cls_name)
    curarg_seps = itertools.chain(filter_compare_ops, filter_combine_ops, (filter_cls.INVERT_CHAR,))
    return Candidates(get_names(filter_cls),
                      curarg_seps=curarg_seps,
                      label=filter_labels[filter_cls_name])

@functools.lru_cache(maxsize=None)
def get_filter_cls(name):
    try:
        return getattr(filter_clses, name)
    except AttributeError:
        raise ValueError('Not a filter class: %r' % name)

@functools.lru_cache(maxsize=None)
def get_filter_names(filter_cls):
    return itertools.chain(filter_cls.BOOLEAN_FILTERS,
                           filter_cls.COMPARATIVE_FILTERS)

@functools.lru_cache(maxsize=None)
def get_filter_spec(filter_cls, name):
    try:
        return filter_cls.BOOLEAN_FILTERS[name]
    except KeyError:
        try:
            return filter_cls.COMPARATIVE_FILTERS[name]
        except KeyError:
            raise ValueError('No such filter: %r' % (name,))

@functools.lru_cache(maxsize=None)
def filter_takes_completable_values(filter_cls, name):
    try:
        filter_spec = get_filter_spec(filter_cls, name)
    except ValueError:
        return False
    else:
        return (name in filter_cls.COMPARATIVE_FILTERS and
                filter_spec is not None and
                issubclass(filter_spec.value_type, str))



def find_subtree(torrent, path):
    """
    `torrent` must be a Torrent instance and `path` must be a sequence of nested
    directory names in the tree `torrent['files']`

    Empty strings in `path` are ignored (like `ls foo//bar///baz` also works).

    If `torrent` is a single-file torrent or any item in `path[:-1]` points to a
    leaf or doesn't exist, return None.  Otherwise, return the subtree or
    file/leaf that `path` points to.  (To clarify: the last item in `path` may
    not exist, in which case its parent tree is returned.)
    """
    tree = torrent['files'][torrent['name']]
    if tree.nodetype == 'leaf':
        # `torrent` is single-file torrent
        return None
    else:
        for i,part in enumerate(path):
            if part:
                subtree = tree.get(part)
                is_last_part = i == len(path) - 1
                if subtree is None:
                    if not is_last_part:
                        return None
                elif subtree.nodetype == 'leaf' and is_last_part:
                    return subtree
                elif subtree.nodetype == 'parent':
                    tree = subtree
                else:
                    # An item in `path[:-1]` points to a leaf.  We can't use
                    # files as subtrees.
                    return None
        return tree

def find_files(tree):
    """Files at TorrentFileTree object `tree`"""
    for item in tree:
        if tree[item].nodetype == 'leaf':
            yield item

def find_dirs(tree):
    """Directories at TorrentFileTree object `tree`"""
    for item in tree:
        if tree[item].nodetype == 'parent':
            yield item



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio

import urwid

from ... import objects
from ...client import SettingFilter
from ...utils.usertypes import Bool, Option
from .base import ItemWidgetBase, ListWidgetBase
from .setting import TUICOLUMNS

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


def _change_setting(name, new_value, on_success=None):
    if objects.cfg.is_local(name):
        try:
            objects.cfg[name] = new_value
        except ValueError as e:
            log.error('Cannot set %s = %r: %s', name, new_value, e)
        else:
            if on_success is not None:
                on_success()

    elif objects.cfg.is_remote(name):
        async def setter():
            try:
                await objects.cfg.set(name, new_value)
            except (ValueError, objects.srvapi.ClientError) as e:
                log.error('Cannot set %s = %r: %s', name, new_value, e)
            else:
                if on_success is not None:
                    on_success()
        asyncio.ensure_future(setter())

    else:
        raise RuntimeError('Not a setting name: %r' % name)


class SettingItemWidget(ItemWidgetBase):
    palette_unfocused = 'settinglist'
    palette_focused   = 'settinglist.focused'
    columns_focus_map = {}
    for col in TUICOLUMNS.values():
        columns_focus_map.update(col.style.focus_map)

    @property
    def id(self):
        return self.data['id']

    def selectable(self):
        return True

    @property
    def name(self):
        if self._cells.exists('name'):
            return self._cells.name.text.text

    @property
    def value_widget(self):
        if self._cells.exists('value'):
            return self._cells.value.base_widget

    @property
    def current_value(self):
        if self._cells.exists('value'):
            value_widget = self.value_widget
            if hasattr(value_widget, 'get_edit_value'):
                return value_widget.get_edit_value()
            elif isinstance(value_widget, urwid.Edit):
                return value_widget.edit_text
            elif hasattr(value_widget, 'get_tui_value'):
                return value_widget.get_tui_value()

    @property
    def edit_mode(self):
        if self._cells.exists('value'):
            return isinstance(self._cells.value.base_widget, urwid.Edit)
        return False

    def keypress(self, size, key):
        current_value = self.current_value
        if current_value is None:
            return key
        elif isinstance(current_value, Bool):
            return self._keypress_bool(size, key)
        elif isinstance(current_value, Option):
            return self._keypress_option(size, key)
        else:
            return self._keypress_string(size, key)

    def _keypress_bool(self, size, key):
        cmd = self._command_map[key]
        if cmd is urwid.ACTIVATE:
            new_value = not self.current_value
            _change_setting(self.name, new_value)
        else:
            return key

    def _keypress_option(self, size, key):
        cmd = self._command_map[key]
        if cmd is urwid.ACTIVATE:
            current_value = self.current_value
            options = current_value.options
            index = options.index(current_value)
            if index < len(options) - 1:
                index += 1
            else:
                index = 0
            new_value = options[index]
            _change_setting(self.name, new_value)
        else:
            return key

    def _keypress_string(self, size, key):
        cells = self._cells
        cmd = self._command_map[key]
        current_value = self.current_value
        value_widget = self.value_widget

        def edit():
            attrmap = self._cells.value.attrmap
            edit_widget = urwid.AttrMap(urwid.Edit(edit_text=str(current_value)),
                                        attr_map=attrmap.attr_map,
                                        focus_map=attrmap.focus_map)
            cells.replace('value', edit_widget)
            self._value_widget_temp = value_widget

        def unedit():
            cells.replace('value', self._value_widget_temp)
            delattr(self, '_value_widget_temp')

        if cmd is urwid.ACTIVATE:
            if not self.edit_mode:
                edit()
            else:
                new_value = value_widget.edit_text
                _change_setting(self.name, new_value, on_success=unedit)
        elif cmd is urwid.CANCEL:
            if self.edit_mode:
                unedit()
        elif self.edit_mode:
            key = super().keypress(size, key)
            cmd = self._command_map[key]
            # Don't allow user to focus next/previous setting when editing
            if cmd not in (urwid.CURSOR_DOWN, urwid.CURSOR_UP):
                return key
        else:
            return key


class SettingListWidget(ListWidgetBase):
    tuicolumns      = TUICOLUMNS
    ListItemClass   = SettingItemWidget
    keymap_context  = 'setting'
    palette_name    = 'settinglist'
    focusable_items = True

    def __init__(self, srvapi, keymap, sort=None, columns=None, title='Settings'):
        super().__init__(srvapi, keymap, columns=columns, sort=sort, title=title)
        self._sort = sort
        self._secondary_filter = None
        objects.localcfg.on_change(self._handle_update)
        objects.remotecfg.on_update(self._handle_update)
        self.refresh()

    def _handle_update(self, *_, **__):
        self._data_dict = objects.cfg.as_dict
        self._invalidate()

    def refresh(self):
        objects.remotecfg.poll()

    @property
    def sort(self):
        return self._sort

    @sort.setter
    def sort(self, sort):
        ListWidgetBase.sort.fset(self, sort)
        self.refresh()

    @property
    def secondary_filter(self):
        return self._secondary_filter

    @secondary_filter.setter
    def secondary_filter(self, setting_filter):
        if setting_filter is None:
            self._secondary_filter = None
        else:
            self._secondary_filter = SettingFilter(setting_filter)
        self._invalidate()

    def _limit_items(self, setting_widgets):
        sfilter = self._secondary_filter
        if sfilter is not None:
            for sw in setting_widgets:
                if not sfilter.match(sw.data):
                    yield sw



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""General hooks that are always needed regardless of the interface"""

from .objects import cmdmgr, localcfg, srvapi
from .utils import convert
from .views.file import COLUMNS as FILE_COLUMNS
from .views.peer import COLUMNS as PEER_COLUMNS
from .views.torrent import COLUMNS as TORRENT_COLUMNS

from .logging import make_logger  # isort:skip
log = make_logger(__name__)


def _pre_run_hook(cmdline):
    # Dirty hack to make "ls -h" work
    # If there is '-h' or '--help' in the arguments, replace it with 'help
    # <cmd>'.  This is dirty but easier than forcing argparse to ignore all
    # other arguments without calling sys.exit().
    if '-h' in cmdline or '--help' in cmdline:
        cmdcls = cmdmgr.get_cmdcls(cmdline[0], interface='ANY')
        if cmdcls is not None:
            if cmdcls.name != 'tab':
                cmdline = ['help', cmdcls.name]
            else:
                # 'tab ls -h' is a little trickier because both 'tab' and 'ls'
                # can have arbitrary additional arguments which we must remove.
                #
                # Find first argument to 'tab' that is also a valid command
                # name.  Preserve all arguments before that.
                tab_args = []
                for arg in cmdline[1:]:
                    if cmdmgr.get_cmdcls(arg, interface='ANY') is not None:
                        return ['tab'] + tab_args + ['help', arg]
                    else:
                        tab_args.append(arg)
                cmdline = ['help', 'tab']
    return cmdline
cmdmgr.pre_run_hook = _pre_run_hook


_BANDWIDTH_COLUMNS = (TORRENT_COLUMNS['rate-up'], TORRENT_COLUMNS['rate-down'],
                      TORRENT_COLUMNS['limit-rate-up'], TORRENT_COLUMNS['limit-rate-down'],
                      PEER_COLUMNS['rate-up'], PEER_COLUMNS['rate-down'], PEER_COLUMNS['rate-est'])
def _set_bandwidth_unit(settings, name, value):
    convert.bandwidth.unit = value
    unit_short = convert.bandwidth.unit
    for column in _BANDWIDTH_COLUMNS:
        column.set_unit(unit_short)
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_bandwidth_unit, name='unit.bandwidth')
_set_bandwidth_unit(localcfg, name='unit.bandwidth', value=localcfg['unit.bandwidth'])  # Init columns' units

def _set_bandwidth_prefix(settings, name, value):
    convert.bandwidth.prefix = value
    for column in _BANDWIDTH_COLUMNS:
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_bandwidth_prefix, name='unitprefix.bandwidth')


_SIZE_COLUMNS = (TORRENT_COLUMNS['size'], TORRENT_COLUMNS['downloaded'],
                 TORRENT_COLUMNS['uploaded'], TORRENT_COLUMNS['available'],
                 FILE_COLUMNS['size'], FILE_COLUMNS['downloaded'])
def _set_size_unit(settings, name, value):
    convert.size.unit = value
    unit_short = convert.size.unit
    for column in _SIZE_COLUMNS:
        column.set_unit(unit_short)
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_size_unit, name='unit.size')
_set_size_unit(localcfg, name='unit.size', value=localcfg['unit.size'])  # Init columns' units

def _set_size_prefix(settings, name, value):
    convert.size.prefix = value
    for column in _SIZE_COLUMNS:
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_size_prefix, name='unitprefix.size')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import operator
from collections import abc

from .. import constants as const
from ..utils import SmartCmpStr
from .base import BoolFilterSpec, CmpFilterSpec, Filter, FilterChain, FilterSpecDict


def _is_sequence(obj):
    return isinstance(obj, abc.Sequence) and not isinstance(obj, str)

# Setting values have multiple incompatible types.  Try hard to compare them to
# the user-given value or fail silently.
def _match_coerced_value(setting, key, op, value):
    setting_value = setting[key]

    # Try to make `value` the same type as `setting['value']`
    validate = setting.get('validate', None)
    if validate is not None:
        try:
            value = validate(value)
        except ValueError:
            return False

    # Sequence matches if all its items are in `setting['value']`
    if _is_sequence(setting_value) and _is_sequence(value):
        if op is operator.__contains__:
            # Match if `settings['value'] contains all items of `value`, in any order
            for v in value:
                if v not in setting_value:
                    return False
            return True
        else:
            try:
                return op(setting_value, value)
            except TypeError:
                return False
    else:
        try:
            return op(setting_value, value)
        except TypeError:
            return False


class _SingleFilter(Filter):
    DEFAULT_FILTER = 'name'

    BOOLEAN_FILTERS = FilterSpecDict({
        'all'     : BoolFilterSpec(None,
                                   aliases=('*',),
                                   description='All settings'),
        'changed' : BoolFilterSpec(lambda s: s['value'] not in (s['default'], const.DISCONNECTED),
                                   aliases=('ch',),
                                   description='Settings with customized values'),
    })

    COMPARATIVE_FILTERS = FilterSpecDict({
        'name'        : CmpFilterSpec(value_getter=lambda s: s['id'],
                                      value_type=SmartCmpStr,
                                      aliases=('n',),
                                      description='Match VALUE against setting name'),
        'value'       : CmpFilterSpec(value_getter=lambda s: s['value'],
                                      value_matcher=lambda s, op, v: _match_coerced_value(s, 'value', op, v),
                                      value_type=SmartCmpStr,
                                      aliases=('v',),
                                      description='Match VALUE against current value'),
        'default'     : CmpFilterSpec(value_getter=lambda s: s['default'],
                                      value_matcher=lambda s, op, v: _match_coerced_value(s, 'default', op, v),
                                      value_type=SmartCmpStr,
                                      aliases=('def',),
                                      description='Match VALUE against default value'),
        'description' : CmpFilterSpec(value_getter=lambda s: s['description'],
                                      value_type=SmartCmpStr,
                                      aliases=('desc',),
                                      description='Match VALUE against description'),
    })


class SettingFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = _SingleFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import re
import textwrap
from shutil import get_terminal_size
from types import SimpleNamespace

from ...utils.string import crop_and_align, normalize_unicode, stralign, strwidth

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


TERMSIZE = get_terminal_size(fallback=(None, None))


_whitespace_regex = re.compile(r'^\s*$')
def _wrapped(line, width):
    # Prevent textwrap.wrap() from filtering out empty lines
    if _whitespace_regex.match(line):
        yield line
    else:
        yield from textwrap.wrap(line, width=width, break_on_hyphens=False)

def _get_cell_lines(cell):
    # Return string of single cell correctly cropped/padded and aligned
    line = normalize_unicode(str(cell.get_cli_value()))
    width = cell.width
    if isinstance(width, int):
        if cell.wrap == 'clip':
            return (crop_and_align(line, width, cell.align,
                                   has_wide_chars=cell.may_have_wide_chars),)
        else:
            return tuple(stralign(line, width=width)
                         for line in _wrapped(line, width))
    else:
        return (line,)

def _assemble_row(table, line_index, pretty=True):
    # Concatenate all cells in a row with delimiters
    # Return a list of lines (cells may have multiple lines)
    row = []
    for cell in table.rows[line_index]:
        if pretty:
            row.append(_get_cell_lines(cell))
        else:
            row.append((str(cell.get_raw_value()),))

    lines_count = max(len(cell) for cell in row)
    lines = []
    delimiter = table.delimiter
    for i in range(lines_count):
        # `row` is a list of cells; each `cell` is a list of lines in an
        # individual cell.
        line = []
        for cell in row:
            cell_lines = len(cell)
            empty_space = ' ' * len(cell[0])
            line.append(cell[i] if i < cell_lines else empty_space)
        line = delimiter.join(line)
        lines.append(line)
    return lines

def _assemble_headers(table):
    # Concatenate all column headers with delimiters
    # This must be called after shrink_and_expand_to_fit() so we can
    # grab the final column widths from the first row.
    headers = []
    for colname in table.colorder:
        width = table.colwidths[colname]
        header_items = table.colspecs[colname].header
        left  = header_items.get('left', '')
        right = header_items.get('right', '')
        space = ' ' * (width - len(left) - len(right))
        header = ''.join((left, space, right))[:width]
        headers.append(header)
    return table.delimiter.join(headers)

def _get_header_width(table, colname):
    header = table.colspecs[colname].header
    return strwidth(' '.join((header.get('left', ''),
                              header.get('right', ''))).strip())

def _get_colwidth(table, colindex):
    # Return width of widest cell in column
    rows = (normalize_unicode(str(row[colindex].get_cli_value()))
            for row in table.rows)
    return max(strwidth(row) for row in rows)

def _column_has_variable_width(table, colname):
    # Whether column has fixed or variable width
    return not isinstance(table.colspecs[colname].width, int)

def _column_could_use_more_width(table, colname):
    # Whether any cell in a column is cropped
    return table.colwidths[colname] < table.maxcolwidths[colname]

def _column_is_shrinkable(table, colname):
    # Whether the current width of column is larger than its minimum width
    return table.colwidths[colname] > table.colspecs[colname].min_width

def _set_colwidth(table, colindex, width):
    # Set width of all cells in a column
    colname = table.colorder[colindex]
    table.colwidths[colname] = width
    for row in table.rows:
        row[colindex].width = width

def _set_maxcolwidth(table, colindex, maxwidth):
    # Set maximum width a column can make use of (no cell needs more horizontal
    # space than `maxwidth`)
    colname = table.colorder[colindex]
    table.maxcolwidths[colname] = maxwidth

def _get_excess_width(table):
    # Return width by which table must be narrowed to fit in max_width
    first_line = _assemble_row(table, 0)[0]
    return strwidth(first_line) - table.max_width

def _remove_column(table, colindex):
    # Delete column from internal structures
    del table.colwidths[table.colorder[colindex]]
    del table.colorder[colindex]
    for row in table.rows:
        del row[colindex]

def _shrink_to_widest_value(table):
    # Reduce width of columns where header and all values are narrower than the
    # current width
    for colindex,colname in enumerate(table.colorder):
        max_value_width = max(_get_header_width(table, colname),
                              _get_colwidth(table, colindex))
        _set_colwidth(table, colindex, max_value_width)
        _set_maxcolwidth(table, colindex, max_value_width)

def _shrink_variable_width_columns(table):
    # Reduce width of columns that haven't reached their min_size yet
    excess = _get_excess_width(table)
    while excess > 0:
        candidates = [(colname,table.colwidths[colname])
                      for colname in table.colorder
                      if _column_is_shrinkable(table, colname)]

        if len(candidates) >= 2:
            # Sort by width (first item is widest)
            candidates_sorted = sorted(candidates, key=lambda col: col[1], reverse=True)
            widest0_name, widest0_width = candidates_sorted[0]
            widest1_name, widest1_width = candidates_sorted[1]
            # Shrink widest column by difference to second widest column
            # (leaving them at the same width), but not by more than `excess`
            # characters.
            shrink_amount = max(1, min(excess, widest0_width - widest1_width))
        elif len(candidates) >= 1:
            # Only one column left to shrink
            widest0_name = candidates[0][0]
            shrink_amount = 1
        else:
            # No shrinkable columns
            break

        new_width = table.colwidths[widest0_name] - shrink_amount
        _set_colwidth(table, table.colorder.index(widest0_name), new_width)
        excess = _get_excess_width(table)

def _shrink_by_removing_columns(table):
    # Remove columns until table is no longer wider than terminal
    while _get_excess_width(table) > 0:
        _remove_column(table, 0)

    # We may have freed up space to give back to columns of variable width
    freed_width = -_get_excess_width(table)
    if freed_width > 0:
        while freed_width > 0:
            freed_width -= 1
            # Find non-fixed-width columns that could use more width
            candidates = [(colname,table.colwidths[colname])
                          for colname in table.colorder
                          if (_column_has_variable_width(table, colname) and
                              _column_could_use_more_width(table, colname))]
            if not candidates:
                # We have space left, but no column wants it
                break
            candidates_sorted = sorted(candidates, key=lambda col: col[1])
            colname = candidates_sorted[0][0]
            new_width = table.colwidths[candidates_sorted[0][0]] + 1
            _set_colwidth(table, table.colorder.index(colname), new_width)

def _fit_table_into_terminal(table):
    # Expand all cells in each column to the width of the widest cell
    # Keep track of each column width in table namespace
    for colindex,colname in enumerate(table.colorder):
        if _column_has_variable_width(table, colname):
            colwidth = _get_colwidth(table, colindex)
        else:
            colwidth = table.colspecs[colname].width
        _set_colwidth(table, colindex, colwidth)

    _shrink_to_widest_value(table)
    _shrink_variable_width_columns(table)
    _shrink_by_removing_columns(table)

def print_table(items, order, column_specs):
    """
    Print table from a two-dimensional array of column objects

    `column_specs` maps column IDs to ColumnBase classes.  A column ID is any
    hashable object, but you probably want strings like 'name', 'id', 'date',
    etc.

    `order` is a sequence of column IDs.

    `items` is a sequence of arbitrary objects that are used to create cell
    objects by passing them to the classes in `column_specs`.
    """
    # Whether to print for a human or for a machine to read our output
    pretty_output = all(x is not None for x in (TERMSIZE.columns, TERMSIZE.lines))

    table = SimpleNamespace(colspecs=column_specs,
                            colorder=order,
                            colwidths={}, maxcolwidths={},  # Calculated when needed
                            delimiter='\t' if TERMSIZE.columns is None else '│',
                            max_width=TERMSIZE.columns)

    # Create two-dimensional list of cells.  Each cell must behave like an
    # instance of a child class of ColumnBase (see stig.views.__init__.py).
    table.rows = []
    for item in items:
        row = []
        for i,colname in enumerate(table.colorder):
            row.append(table.colspecs[colname](item))
        table.rows.append(row)

    if len(table.rows) > 0:
        if pretty_output:
            _fit_table_into_terminal(table)
            headerstr = '\033[1;4m' + _assemble_headers(table) + '\033[0m'
        else:
            log.debug('Could not detect TTY size - assuming stdout is no TTY')

        for line_index in range(len(table.rows)):
            # Print column headers after every screen full
            if pretty_output and line_index % (TERMSIZE.lines - 2) == 0:
                print(headerstr)
            for row in _assemble_row(table, line_index, pretty=pretty_output):
                print(row)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import argparse
import asyncio
import sys
from collections import abc

from ..completion import Candidate, Candidates
from ..utils import cliparser
from .cmderror import CmdArgError, CmdError

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class StayAliveArgParser(argparse.ArgumentParser):
    """Capture errors instead of printing them on stderr and exiting"""
    def exit(self, status=0, message=None):
        if message is not None:
            self.error(message)

    def error(self, message):
        raise CmdArgError(message[0].upper() + message[1:])


_MANDATORY_CMD_ATTRS = ('name', 'category', 'provides', 'description', 'run')
_OPTIONAL_CMD_ATTRS = {
    'aliases': (), 'usage': (), 'examples': (), 'argspecs': (), 'more_sections': {}
}
def CommandMeta(clsname, bases, attrs):
    """
    Class factory that inits all commands

    This takes care of inheritance, complains about missing mandatory class
    attributes, adds defaults for optional class attributes, etc.
    """
    # Make sure all mandatory attributes are there
    for attr in _MANDATORY_CMD_ATTRS:
        if attr not in attrs:
            raise RuntimeError('{} misses mandatory attribute: {!r}'
                               .format(clsname, attr))

    # Add defaults for optional attributes
    for k,v in _OPTIONAL_CMD_ATTRS.items():
        if k not in attrs:
            attrs[k] = v

    # 'names' attribute
    attrs['names'] = list(attrs['aliases'])
    attrs['names'].insert(0, attrs['name'])

    # Options are arguments that start with '-'
    # Map '--option' to ('-o', '-p', '-t')
    attrs['long_options'] = {name:argspec['names'][1:]
                             for argspec in attrs['argspecs']
                             for name in argspec['names']
                             if name.startswith('--')}
    # Map '-o' to '--option'
    attrs['short_options'] = {name:argspec['names'][0]
                              for argspec in attrs['argspecs']
                              for name in argspec['names']
                              if name.startswith('-') and not name[1:].startswith('-')}

    # Create argument parser
    argp = StayAliveArgParser(prog=attrs['name'], add_help=False)
    for argspec in attrs['argspecs']:
        # Check if all mandatory keys are in the spec
        for mandatory_key in ('names',):
            if mandatory_key not in argspec:
                raise RuntimeError('Missing key {!r} in argument spec: {}'
                                   .format(mandatory_key, argspec))

        # Create a copy of argspec so we don't alter the original class
        # attribute and remove all items that ArgParser doesn't understand
        argspec = argspec.copy()
        if 'description' in argspec:
            argspec.pop('description')
        if 'default_description' in argspec:
            argspec.pop('default_description')
        if 'document_default' in argspec:
            argspec.pop('document_default')

        # Assemble arg names
        argnames = argspec.pop('names', None)
        if not isinstance(argnames, abc.Sequence):
            argnames = (argnames,)

        # Translate string to argparse constant
        if 'nargs' in argspec and argspec['nargs'] == 'REMAINDER':
            argspec['nargs'] = argparse.REMAINDER

        # Create new argparser
        argp.add_argument(*argnames, **argspec)
    attrs['_argparser'] = argp

    # Command class must inherit from _CommandBase
    if _CommandBase not in bases:
        bases = (_CommandBase, _CompletionCandidatesMixin) + bases

    cls = type(clsname, bases, attrs)
    return cls


class _CommandBase():
    """
    Base for all command classes

    args: Any arguments for this command
    argv: Command line as given by user, including command name, as sequence
    info_handler: Function that is called with info messages via the info() method
    error_handler: Function that is called with error messages via the error() method

    Any remaining keyword arguments are made available as instance attributes.
    """
    def __init__(self, args=(), argv=(), info_handler=None, error_handler=None, **kwargs):
        # Store the command that invoked this instance
        self._command = ' '.join((cliparser.quote(arg) for arg in argv))
        self._args = args
        for k,v in kwargs.items():
            setattr(self, k, v)
        self._info_handler = info_handler or (lambda msg: print(msg, file=sys.stdout))
        self._error_handler = error_handler or (lambda msg: print(msg, file=sys.stderr))
        self._task = None
        self._success = None
        self._is_async = False

        try:
            args_parsed = self._argparser.parse_args(args)
        except CmdArgError as e:
            self._finish(exception=e)
        else:
            # Create keyword args for run() method
            kwargs = {}
            for argspec in self.argspecs:
                # First name is the kwarg for run()
                key = argspec.get('dest') or argspec['names'][0].lstrip('-').replace('-', '_')
                value = getattr(args_parsed, key)
                kwargs[key.replace(' ', '_')] = value
            self._args = kwargs

            if asyncio.iscoroutinefunction(self.run):
                log.debug('Running async command: %r', self)
                self._is_async = True
                self._task = asyncio.ensure_future(self.run(**kwargs))
                self._task.add_done_callback(lambda task: self._catch_exceptions(task.result))
            else:
                log.debug('Running sync command: %r', self)
                self._is_async = False
                self._catch_exceptions(self.run, **kwargs)

    def _catch_exceptions(self, callabee, *args, **kwargs):
        try:
            callabee(*args, **kwargs)
        except Exception as e:
            self._finish(exception=e)
        else:
            self._finish()

    def _finish(self, exception=None):
        if not self.finished:
            log.debug('Finishing %s', self)
            self._success = not bool(exception)
            if isinstance(exception, CmdError):
                exc_str = str(exception)
                if exc_str:
                    self.error(exc_str)

        if exception and not isinstance(exception, CmdError):
            raise exception

    def wait_sync(self):
        """
        Wait synchronously until this command has finished

        This uses the run_until_complete() method, so the loop must not be
        running or closed.
        """
        if not self.finished:
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                raise RuntimeError('AsyncIO loop is closed - cannot wait for {!r}'.format(self))
            elif loop.is_running():
                raise RuntimeError('AsyncIO loop is running - cannot wait for {!r}'.format(self))
            else:
                log.debug('Waiting until finished: %r', self)
                self._catch_exceptions(loop.run_until_complete, self._task)

    async def wait_async(self):
        """Wait asynchronously until this command has finished"""
        if not self.finished:
            log.debug('Waiting until finished: %r', self)
            try:
                await self._task
            except Exception as e:
                self._finish(exception=e)
            else:
                self._finish()

    def info(self, msg):
        """Show info message (use this as your stdout)"""
        self._info_handler('%s: %s' % (self.name, msg))

    def error(self, msg):
        """Show error message (use this as your stderr)"""
        self._error_handler('%s: %s' % (self.name, msg))

    @property
    def success(self):
        """True/False if command has finished, None otherwise"""
        return self._success

    @property
    def finished(self):
        """Whether command has finished"""
        return self._success is not None

    @property
    def is_async(self):
        """Whether this command runs asynchronously"""
        return self._is_async

    @property
    def command(self):
        """Invoking command as string"""
        return self._command

    def __repr__(self):
        if isinstance(self._args, abc.Sequence):
            argstr = ', '.join(repr(arg) for arg in self._args)
        elif isinstance(self._args, abc.Mapping):
            argstr = ', '.join('%s=%r' % (k,v)
                               for k,v in self._args.items())
        provides = '/'.join(interface for interface in self.provides)
        string = '<Command [{}] {}({})'.format(provides, self.name, argstr)
        if self.finished:
            string += ' success={}'.format(self.success)
        else:
            string += ' running'
        return string + '>'


class _CompletionCandidatesMixin():
    """
    Mixin class that provides class methods for generating tab completion
    candidates
    """

    @classmethod
    def _is_option(cls, arg):
        return arg in cls.long_options or arg in cls.short_options

    @classmethod
    def _get_argspec(cls, name):
        for argspec in cls.argspecs:
            if any(name == n for n in argspec['names']):
                return argspec

    @classmethod
    def _option_wants_arg(cls, option, roffset):
        argspec = cls._get_argspec(option)
        nargs = argspec.get('nargs', 1)
        if nargs in ('+', '*', 'REMAINDER'):
            return True
        elif nargs == '?':
            nargs = 1
        if roffset < nargs:
            return True
        return False

    @classmethod
    def completion_candidates_opts(cls, args):
        """
        Return candidates for arguments that start with "-"

        The default implementation should work for all commands and it shouldn't
        be necessary for subclasses to override this method.
        """
        # '--' turns all arguments to the right into positional arguments
        if '--' not in args.before_curarg:
            if args.curarg.startswith('-'):
                log.debug('Completing long options: %r', cls.long_options)
                # Remove any options from candidates that are already in `args`
                options = tuple(opt for opt in cls.long_options if opt not in args)
                if options:
                    cands = (Candidate(cand,
                                       in_parens=', '.join(cls.long_options[cand]),
                                       description=cls._get_argspec(cand)['description'])
                             for cand in options)
                    return Candidates(cands, label='%s option' % cls.name)

            # Check if any argument left of the current argument is an option that
            # wants another parameter
            for i,arg in enumerate(reversed(args.before_curarg)):
                if cls._is_option(arg) and cls._option_wants_arg(option=arg, roffset=i):
                    option_name = cls.short_options.get(arg, arg)
                    log.debug('Completing parameters for %r', option_name)
                    return cls.completion_candidates_params(option_name, args)

    @classmethod
    def completion_candidates_params(cls, option_name, args):
        """
        Return candidates for arguments to an option (e.g. --option parameter)

        The default implementation returns None.
        """
        pass

    @classmethod
    def completion_candidates_posargs(cls, args):
        """
        Return candidates that are not options and not parameters

        The default implementation returns None.
        """
        pass

    @classmethod
    def completion_candidates(cls, args):
        """
        Return appropriate completion candidates

        The default implementation calls the methods described above and returns
        the first return value that evaluates to True (i.e. non-empty and
        non-None).
        """
        # completion_candidates_params() is called by completion_candidates_opts
        # if appropriate.
        cands = cls.completion_candidates_opts(args)
        log.debug('Completion candidates for options: %r', cands)
        if cands:
            return cands
        cands = cls.completion_candidates_posargs(args)
        log.debug('Completion candidates for positional arguments: %r', cands)
        return cands



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Anything related to the help system that is common between interfaces"""

import re
import string
from collections import abc

from . import __appname__, __version__, objects
from .cliopts import DESCRIPTIONS as CLI_DESCRIPTIONS
from .utils import expandtabs
from .utils.string import striplines
from .utils.usertypes import Float, Int

from .logging import make_logger  # isort:skip
log = make_logger(__name__)


class ForgivingFormatter(string.Formatter):
    def get_value(self, key, args, kwargs):
        if isinstance(key, str):
            try:
                return kwargs[key]
            except KeyError:
                return '{%s}' % key
        else:
            return super().get_value(key, args, kwargs)

    def __call__(self, lines):
        return tuple(
            striplines(self.format(line, __appname__=__appname__)
                       for line in expandtabs.expand(lines, indent=2))
        )

finalize_lines = ForgivingFormatter()


class HelpManager():
    """Provide help texts for CommandManager, Settings and KeyMap objects"""

    MAIN_TOPICS = {
        'commandsmanual' : 'Describes how to call and chain commands',
        'commands'       : 'Lists commands',
        'filtersmanual'  : 'Describes how to define and combine filters',
        'filters'        : 'Lists filters for torrents, files, etc',
        'settingsmanual' : 'Describes how to change settings',
        'settings'       : 'Lists configuration settings',
        'keybindings'    : 'Lists TUI keybindings',
    }

    ALIASES = {
        'cmds'        : 'commands',
        'cmdsman'     : 'commandsmanual',
        'filtersman'  : 'filtersmanual',
        'config'      : 'settings',
        'cfg'         : 'settings',
        'configman'   : 'settingsmanual',
        'cfgman'      : 'settingsmanual',
        'keymap'      : 'keybindings',
        'keys'        : 'keybindings',
    }

    def find(self, topic=None):
        """Find help topic and return lines"""
        if topic in self.ALIASES:
            topic = self.ALIASES[topic]

        if topic is None:
            return self.topic_overview

        if hasattr(self, 'topic_' + topic):
            return getattr(self, 'topic_' + topic)
        elif topic in objects.cmdmgr:
            return self.command(topic)
        elif topic in objects.cfg:
            return self.setting(topic)

        raise ValueError('Unknown help topic: %r' % topic)

    @property
    def topic_overview(self):
        lines = [
            '{} {}'.format(__appname__, __version__),
            '',
            'SYNTAX',
            '\t{__appname__} [<OPTIONS>] [<COMMANDS>]',
            '',
        ]

        for section,opts in CLI_DESCRIPTIONS.items():
            lines.append('%s' % section.upper())
            for opts,desc in opts.items():
                lines.append('\t%s  \t%s' % (opts, desc))
            lines.append('')

        def topic_line(topic):
            names = (topic,) + tuple(alias for alias,topic_ in self.ALIASES.items()
                                     if topic_ == topic)
            return '\t\t%s \t- \t%s' % (', '.join(names), self.MAIN_TOPICS[topic])

        lines += ['HELP TOPICS',
                  ('\tAll commands and settings are valid help topics.  Read '
                   "them with '{__appname__} help <TOPIC>' or '{__appname__} -h <TOPIC>'.  "
                   'Additionally, the following topics are available:')]
        lines.extend(topic_line(topic) for topic in self.MAIN_TOPICS)

        return finalize_lines(lines)

    @property
    def topic_settingsmanual(self):
        lines = [
            'SETTINGS',
            ("\tSettings can be changed with the commands 'set' and 'reset' "
             "(see 'help set' and 'help reset')."),
            '',
            ('\tLocal settings change the behaviour of {__appname__} while '
             'remote settings change the behaviour of the connected daemon.  '
             'The names of remote settings start with "srv.".'),
            '',
            ('\tChanges made to local settings are not permanent.  All '
             'values are set back to their defaults once {__appname__} is '
             'restarted (see RC FILES).'),
            '',
            ('\tChanges made to remote settings, on the other hand, are '
             'permanent as the daemon has its own configuration system.'),
            '',
            ('\tIn the TUI, the "set" command with no arguments (bound to <alt-s> '
             'by default) lists all settings and lets you edit them with <enter>.  '
             'The "dump" command (bound to <alt-S> by default) makes your '
             'current settings, keybindings and tabs permanent.'),
            '',
            'RC FILES',
            ('\tAn rc file contains a list of arbitrary commands.  '
             r'Commands can span multiple lines by escaping line breaks with "\".  '
             'Lines starting with "#" (optionally preceded by spaces) are ignored.'),
            '',
            ('\tCommands in an rc file are called during startup before the '
             'commands given on the command line.'),
            '',
            ('\tThe default rc file path is "$XDG_CONFIG_HOME/{__appname__}/rc", '
             'where $XDG_CONFIG_HOME defaults to "~/.config" if it is not set.'),
            '',
            ('\tA different path can be provided with the "--rcfile" option.  '
             'An existing rc file at the default path can be ignored with the '
             '"--norcfile" option.'),
            '',
            '\tTo permanently change the default config file, create an alias:',
            '',
            '\t\t$ alias {__appname__}="command {__appname__} --rcfile ~/.{__appname__}rc"',
            '',
            ('\tTo load any additional rc files after the default one use the '
             '"rc" command.  (Note that this will prevent the TUI from being '
             'loaded unless you provide the "--tui" option.  See the GUESSING '
             'THE USER INTERFACE section in the "commandsmanual" help for '
             'more information).'),
            '',
            ('\tTUI commands (e.g. "tab" or "bind") in an rc file are ignored '
             'in CLI mode.'),
        ]

        return finalize_lines(lines)

    @property
    def topic_settings(self):
        """Return help text for all settings"""
        localcfg = objects.localcfg
        remotecfg = objects.remotecfg

        lines = []
        lines.append('LOCAL SETTINGS')
        for name in sorted(localcfg):
            lines.append('\t' + name + '  \t' + localcfg.description(name))

        lines += ['']

        lines.append('REMOTE SETTINGS')
        for name in sorted(remotecfg):
            lines.append('\t' + name + '  \t' + remotecfg.description(name))
        return finalize_lines(lines)

    def setting(self, name):
        """Return help text for setting"""
        cfg = objects.cfg
        if name not in objects.cfg:
            raise ValueError('Unknown help topic: %r' % name)
        value = cfg[name]

        def pretty_string(value):
            if isinstance(value, str) and re.match(r'^\s+$', value):
                return repr(value)
            elif isinstance(value, (Float, Int)):
                return value.with_unit
            else:
                return str(value)

        lines = ['%s - \t%s' % (name, cfg.description(name)),
                 '\tValue: \t' + pretty_string(cfg[name]),
                 '\tDefault: \t' + pretty_string(cfg.default(name))]

        if hasattr(value, 'options'):
            opt_strs = []
            for opt in sorted(value.options):
                opt_strs.append(str(opt))
                aliases = [alias for alias,option in value.aliases.items()
                           if option == opt]
                if aliases:
                    opt_strs[-1] += ' (%s)' % (', '.join(aliases))
            lines.append('\tOptions: \t' + ', '.join(opt_strs))

        lines.append('\tSyntax: \t' + cfg.syntax(name))

        return finalize_lines(lines)

    @property
    def topic_commandsmanual(self):
        from .commands import OPS_AND, OPS_OR, OPS_SEQ
        lines = [
            'COMMANDS',
            '\tCommands can be called:',
            '\t\t- \tby providing them as command line arguments,',
            "\t\t- \tvia the command line in the TUI (press ':' to open it),",
            "\t\t- \tby binding them to keys (see 'help bind'),",
            ("\t\t- \tby listing them in an rc file (see 'help cfgman') "
             "and loading it with the '--rcfile' option or the 'rc' command."),
            '',
            'CHAINING COMMANDS',
            ("\tCombining commands with operators makes it possible to run "
             "a command based on the previous command's success."),
            "",
            "\tAvailable command operators are: ",
            "\t\t%s \t- \tRun the next command if the previous command succeeded." % '/'.join(OPS_AND),
            "\t\t%s \t- \tRun the next command if the previous command failed." % '/'.join(OPS_OR),
            "\t\t%s \t- \tRun the next command in any case." % '/'.join(OPS_SEQ),
            "",
            "\tCommand operators must be enclosed by spaces.",
            "",
            ("\tFor example, 'ls foo & ls bar' would list all 'foo' torrents and, "
             "if any where found, continue to list all 'bar' torrents.  "
             "However, 'ls foo | ls bar' would list 'bar' torrents only if there "
             "are no 'foo' torrents."),
            '',
            'GUESSING THE USER INTERFACE (CLI/TUI)',
            ("\tIf commands are given as command line arguments and neither "
             "'--tui' nor '--notui' are provided, {__appname__} tries to guess "
             "whether it makes sense to start the TUI or just run the commands "
             "and exit.  For example, if you run '{__appname__} stop foo', "
             "it is reasonable to assume that you want to run 'stop foo' and "
             "get your shell prompt back.  But if you run "
             "'{__appname__} set connect.host foo.bar', "
             "you probably expect the TUI to pop up."),
            '',
            "\tThis is how this works basically:",
            ("\t\t- \tWithout CLI commands, the TUI is loaded and vice versa."),
            ("\t\t- \tCommands in the torrent category (see 'help commands') prevent the TUI."),
            ("\t\t- \tChanging TUI settings ('(re)set tui.*') enforces the TUI."),
            ("\t\t- \tChanging remote settings ('set srv.*') prevents the TUI."),
            ("\t\t- \tCommands that are exclusive to TUI or CLI (e.g. 'tab') enforce their "
             "interface.  Providing both TUI- and CLI-only commands produces an error.  "
             "Provide --tui or --notui in that case."),
        ]
        return finalize_lines(lines)

    @property
    def topic_commands(self):
        """Must be set to a CommandManager object; provides a help text"""
        lines = []
        for category in objects.cmdmgr.categories:
            lines.append('{} COMMANDS'.format(category.upper()))

            # To deduplicate commands with the same name that provide
            # different interfaces (but should have the same docs), map
            # command names to commands.
            cmds = {}
            for cmd in objects.cmdmgr.all_commands:
                if category == cmd.category:
                    cmds[cmd.name] = cmd

            for cmdname,cmd in sorted(cmds.items()):
                lines.append('\t{}  \t{}'.format(', '.join((cmd.name,) + cmd.aliases),
                                                 cmd.description))
            lines.append('')
        return finalize_lines(lines)

    def command(self, name):
        """Return help text for command"""
        cmd = objects.cmdmgr[name]

        def takes_value(argspec):
            if argspec.get('action') in ('store_true', 'store_false', 'store_const'):
                return False  # Boolean option
            if 'nargs' not in argspec:
                return True
            nargs = argspec['nargs']
            return not isinstance(nargs, int) or nargs > 0

        def arg_dest(argspec):
            if 'metavar' in argspec:
                dest = argspec['metavar'].upper()
            elif 'dest' in argspec:
                dest = argspec['dest'].upper()
            elif argspec['names'][0][0] == '-':
                dest = argspec['names'][0].lstrip('-').upper()
            else:
                return None

            if 'nargs' in argspec and argspec['nargs'] in ('*', '?'):
                return '[<%s>]' % dest
            else:
                return '<%s>' % dest

        lines = [cmd.name.upper()]

        log.debug('Generating help text for %s', cmd.name)
        names = ', '.join((cmd.name,) + cmd.aliases)
        lines = [names + ' - \t' + cmd.description]
        lines.append('')

        if cmd.usage:
            lines.append('USAGE')
            lines += ['\t' + u for u in cmd.usage]
            lines.append('')

        if cmd.argspecs:
            lines.append('ARGUMENTS')
            lines_args = []
            for argspec in cmd.argspecs:
                if 'description' not in argspec:
                    # Argument has no description
                    continue

                arglines = ['\t' + ', '.join(argspec['names'])]
                if takes_value(argspec):
                    dest = arg_dest(argspec)
                    if dest is not None:
                        arglines[0] += ' %s' % dest

                if isinstance(argspec['description'], str):
                    arglines[0] += '  \t' + argspec['description']
                else:  # Assume description is a sequence
                    arglines[0] += '  \t' + argspec['description'][0]
                    for paragraph in argspec['description'][1:]:
                        arglines.append('\t  \t' + paragraph)

                if 'document_default' not in argspec or argspec['document_default']:
                    # Argument takes a value that may default to another value
                    # if omitted and we want to document that default value
                    def stringify_default(default):
                        dflt = default() if callable(default) else default

                        if not isinstance(dflt, str) and isinstance(dflt, abc.Sequence):
                            return ' '.join(dflt)
                        else:
                            return str(dflt)

                    if 'default_description' in argspec:
                        arglines.append('\t  \tDefault: %s' % stringify_default(argspec['default_description']))
                    elif 'default' in argspec:
                        arglines.append('\t  \tDefault: %s' % stringify_default(argspec['default']))

                lines_args.extend(arglines)
            lines += lines_args
            lines.append('')

        if cmd.examples:
            lines.append('EXAMPLES')
            lines += ['\t' + e for e in cmd.examples]
            lines.append('')

        for name,text in sorted(cmd.more_sections.items()):
            lines.append(name.upper())
            if callable(text):
                text = text()
            lines += ['\t' + line for line in text]
            lines.append('')

        return finalize_lines(lines)

    @property
    def topic_keybindings(self):
        """Must be set to a KeyMap object; provides a help text"""
        from .tui.tuiobjects import keymap

        def stringify(s):
            return ' '.join(s) if not isinstance(s, str) else s

        lines = []
        for context in sorted(keymap.contexts, key=lambda c: '' if c is None else c):
            if context is None:
                lines.append('GENERAL KEYBINDINGS')
            else:
                lines.append('{} KEYBINDINGS'.format(context.upper()))

            km = ((key, stringify(action)) for key,action in keymap.map(context))

            # Sort by command
            from natsort import natsorted, ns
            for key,action in natsorted(km, key=lambda pair: pair[1], alg=ns.IGNORECASE):
                if len(action) < 40:
                    lines.append('\t%s  \t%s  \t%s' % (key, action, keymap.get_description(key, context)))
                else:
                    lines.append('\t%s  \t%s' % (key, action))
                    lines.append('\t  \t%s' % (keymap.get_description(key, context),))
            lines.append('')
        return finalize_lines(lines)

    @property
    def topic_filtersmanual(self):
        lines = [
            'FILTERING TORRENTS, FILES, PEERS, ETC',
            ('\tCommands that accept FILTER arguments are applied to items '
             'that match these filters.'),
            '',
            '\tThere are two kinds of filters:',
            '\t\t- \tBoolean filters stand on their own (e.g. "downloading")',
            '\t\t- \tComparative filters need a value (e.g. "seeds>20")',
            '',
            '\tThe syntax of comparative filters is: [[<FILTER NAME>]<OPERATOR>]<VALUE>',
            '',
            ('\tBesides the usual operators (=, !=, >, <, >=, <=), "~" matches if the '
             'item\'s value contains the literal string VALUE and "=~" matches if the '
             'item\'s value matches against the Perl-style regular expression VALUE.'),
            '\tExample: "name~foo" matches all torrents with "foo" in their name.',
            '',
            ('\tIf FILTER NAME is omitted, it defaults to a comparative filter that '
             "makes sense, e.g. \"name\" for torrents (see 'help filters').  "
             'If OPERATOR is omitted, it defaults to "~".'),
            '\tExample: "foo" is the same as "~foo" is the same as "name~foo".',
            '',
            ('\tSpaces at the start and the end of VALUE are always removed.  '
             'If the result starts and ends with matching single or double quotes, the '
             'quotes are removed.  Any other quotes are not interpreted, i.e. they must '
             'not be escaped.'),
            '\tExample: "name = foo " matches "foo"; "name = \' foo \' " matches " foo "',
            '',
            '\tAll filters can be inverted by prepending "!" to the filter name.',
            ('\tExample: "name!=foo" is the same as "!name=foo"; '
             '"!name!=foo" is the same as "name=foo".'),
            '',
            ('\tMatching strings is case-insensitive if VALUE does not contain any '
             'upper-case characters, otherwise it is case-sensitive.'),
            '',
            ('\tWhen matching numbers, the unit prefixes "k", "M", "G", "T" and '
             'their binary counterparts "Ki", "Mi", "Gi", "Ti" are supported.  '
             'The case of unit prefixes is ignored.'),
            ('\tExample: "size>1mi" is the same as "size>1048576" (1 Mebibyte); '
             '"size>1m" is the same as "size>1000000" (1 Megabyte)'),
            '',
            ('\tFor time-based filters, VALUE is either an absolute time stamp '
             'or a relative time delta based on the current time.'),
            '',
            ('\tTime stamps support a date in the format [[YYYY-]MM-]DD or YYYY[-MM] '
             'and a time in the format HH:MM[:SS].  Date and time can be combined by '
             'separating them with a space.'),
            '\tExamples: \t"added=2015-05" \tmatches torrents that were added in May 2015.',
            ('\t\t"completed>=01" \tmatches torrents that finished downloading earlier this month '
             '("01" being the first day of the current month).'),
            ('\t\t"activity<10-17 18:45" \tmatches torrents that were last active before '
             '18:45 (6:45 p.m.) on the 17th of October of this year.'),
            '',
            ('\tTime deltas use the format [in |+|-]AMOUNT[s|m|h|d|w|M|y][ ago].  '
             'The words "in" and "ago" are aliases for "+" and "-".  Negative time '
             'deltas match time stamps in the past and positive time deltas '
             'match time stamps in the future.  Filters have individual defaults for '
             'the sign; e.g. "eta > 1h" is the same as "eta > in 1h" while '
             '"completed > 1h" is the same as "completed > 1h ago".'),
            '',
            ('\tFilters can be combined with the operators "&" (logical AND) '
             'and "|" (logical OR).  Multiple FILTER arguments are implicitly '
             'combined with "|".'),
            '\tExample: "name=foo paused" is the same as "name=foo|paused".',
            '',
            ('\tOperators can be escaped with a preceding "\\" to remove their meaning.'),
            '\tExample: "name=foo\\&bar" matches torrents with the name "foo&bar".',
        ]
        return finalize_lines(lines)

    @property
    def topic_filters(self):
        """Provide help text for arguments to TorrentFilter"""
        from .client import (FileFilter, PeerFilter, SettingFilter, TorrentFilter,
                             TrackerFilter)
        lines = []
        for caption,filt in (('TORRENT FILTERS', TorrentFilter),
                             ('FILE FILTERS', FileFilter),
                             ('PEER FILTERS', PeerFilter),
                             ('TRACKER FILTERS', TrackerFilter),
                             ('SETTING FILTERS', SettingFilter)):
            lines += ['',
                      '%s' % caption,
                      '\tDEFAULT FILTER: %s' % filt.DEFAULT_FILTER,
                      '']

            lines.append('\tBOOLEAN FILTERS')
            for fname,f in sorted(filt.BOOLEAN_FILTERS.items()):
                lines.append('\t\t{} \t{}'.format(', '.join((fname,) + f.aliases), f.description))

            lines += ['', '\tCOMPARATIVE FILTERS']
            for fname,f in sorted(filt.COMPARATIVE_FILTERS.items()):
                if fname == filt.DEFAULT_FILTER:
                    lines.append('\t\t{} \t{} (default)'.format(', '.join((fname,) + f.aliases), f.description))
                else:
                    lines.append('\t\t{} \t{}'.format(', '.join((fname,) + f.aliases), f.description))

        return finalize_lines(lines)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from natsort import humansorted

from ... import objects
from .. import CmdError
from ..base import peer as base
from . import _mixin as mixin
from ._table import print_table

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class ListPeersCmd(base.ListPeersCmdbase,
                   mixin.make_request, mixin.select_torrents):
    provides = {'cli'}

    async def make_peer_list(self, tfilter, pfilter, sort, columns):
        response = await self.make_request(
            objects.srvapi.torrent.torrents(tfilter, keys=('name', 'peers')),
            quiet=True)
        torrents = response.torrents

        if len(torrents) < 1:
            raise CmdError()

        if pfilter is None:
            def filter_peers(peers):
                return peers
        else:
            def filter_peers(peers):
                return pfilter.apply(peers)

        peerlist = []
        for torrent in humansorted(torrents, key=lambda t: t['name']):
            peerlist.extend(filter_peers(torrent['peers']))

        # Pre-lookup peers' IPs
        if 'host' in columns and objects.localcfg['reverse-dns']:
            from ...client import rdns
            rdns.query(*(p['ip'] for p in peerlist))

        sort.apply(peerlist, inplace=True)

        if peerlist:
            from ...views.peer import COLUMNS as PEER_COLUMNS
            print_table(peerlist, columns, PEER_COLUMNS)
        else:
            def filter_is_relevant(f):
                return f and str(f) != 'all'

            if filter_is_relevant(pfilter):
                errmsg = 'No matching peers'
            else:
                errmsg = 'No peers'

            if filter_is_relevant(tfilter):
                errmsg += ' in {} torrents'.format(tfilter)

            if filter_is_relevant(pfilter):
                errmsg += ': {}'.format(pfilter)

            raise CmdError(errmsg)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering TrackerList items by various values"""

from ..ttypes import TorrentTracker
from .base import BoolFilterSpec, CmpFilterSpec, Filter, FilterChain, FilterSpecDict
from .utils import cmp_timestamp_or_timdelta, timestamp_or_timedelta


class _BoolFilterSpec(BoolFilterSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, needed_keys=('trackers',), **kwargs)

class _CmpFilterSpec(CmpFilterSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, needed_keys=('trackers',), **kwargs)

class _SingleFilter(Filter):
    DEFAULT_FILTER = 'domain'

    BOOLEAN_FILTERS = FilterSpecDict({
        'all'   : _BoolFilterSpec(None,
                                  aliases=('*',),
                                  description='All trackers'),
        'alive' : _BoolFilterSpec(lambda trk: trk['status'] != 'stopped',
                                  description='Trackers we are trying to connect to'),
    })

    COMPARATIVE_FILTERS = FilterSpecDict({
        'tier'           : _CmpFilterSpec(value_getter=lambda trk: trk['tier'],
                                         value_type=TorrentTracker.TYPES['tier'],
                                         as_bool=lambda trk: True,
                                         description='Match VALUE against tracker tier'),
        'domain'         : _CmpFilterSpec(value_getter=lambda trk: trk['domain'],
                                         value_type=TorrentTracker.TYPES['domain'],
                                         aliases=('dom', 'tracker'),
                                         description='Match VALUE against domain of announce URL'),
        'url-announce'   : _CmpFilterSpec(value_getter=lambda trk: trk['url-announce'],
                                         value_type=TorrentTracker.TYPES['url-announce'],
                                         aliases=('an',),
                                         description='Match VALUE against announce URL'),
        'url-scrape'     : _CmpFilterSpec(value_getter=lambda trk: trk['url-scrape'],
                                         value_type=TorrentTracker.TYPES['url-scrape'],
                                         aliases=('sc',),
                                         description='Match VALUE against scrape URL'),
        'status'         : _CmpFilterSpec(value_getter=lambda trk: trk['status'],
                                         value_type=TorrentTracker.TYPES['status'],
                                         aliases=('st',),
                                         description=('Match VALUE against tracker status '
                                                      '(stopped, idle, queued, announcing, scraping)')),
        'error'          : _CmpFilterSpec(value_getter=lambda trk: trk['error'],
                                         value_type=TorrentTracker.TYPES['error'],
                                         aliases=('err',),
                                         description='Match VALUE against error message from tracker'),
        'downloads'      : _CmpFilterSpec(value_getter=lambda trk: trk['count-downloads'],
                                         value_type=TorrentTracker.TYPES['count-downloads'],
                                         aliases=('dns',),
                                         description='Match VALUE against number of known downloads'),
        'leeches'        : _CmpFilterSpec(value_getter=lambda trk: trk['count-leeches'],
                                         value_type=TorrentTracker.TYPES['count-leeches'],
                                         aliases=('lcs',),
                                         description='Match VALUE against number of known downloads'),
        'seeds'          : _CmpFilterSpec(value_getter=lambda trk: trk['count-seeds'],
                                         value_type=TorrentTracker.TYPES['count-seeds'],
                                         aliases=('sds',),
                                         description='Match VALUE against number of known seeding peers'),
        'last-announce'  : _CmpFilterSpec(value_getter=lambda trk: trk['time-last-announce'],
                                         value_matcher=lambda trk, op, v: cmp_timestamp_or_timdelta(trk['time-last-announce'], op, v),
                                         value_type=TorrentTracker.TYPES['time-last-announce'],
                                         value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
                                         aliases=('lan',),
                                         description='Match VALUE against time of last announce'),
        'next-announce'  : _CmpFilterSpec(value_getter=lambda trk: trk['time-next-announce'],
                                         value_matcher=lambda trk, op, v: cmp_timestamp_or_timdelta(trk['time-next-announce'], op, v),
                                         value_type=TorrentTracker.TYPES['time-next-announce'],
                                         value_convert=lambda v: timestamp_or_timedelta(v, default_sign=1),
                                         aliases=('nan',),
                                         description='Match VALUE against time of next announce'),
        'last-scrape'    : _CmpFilterSpec(value_getter=lambda trk: trk['time-last-scrape'],
                                         value_matcher=lambda trk, op, v: cmp_timestamp_or_timdelta(trk['time-last-scrape'], op, v),
                                         value_type=TorrentTracker.TYPES['time-last-scrape'],
                                         value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
                                         aliases=('lsc',),
                                         description='Match VALUE against time of last scrape'),
        'next-scrape'    : _CmpFilterSpec(value_getter=lambda trk: trk['time-next-scrape'],
                                         value_matcher=lambda trk, op, v: cmp_timestamp_or_timdelta(trk['time-next-scrape'], op, v),
                                         value_type=TorrentTracker.TYPES['time-next-scrape'],
                                         value_convert=lambda v: timestamp_or_timedelta(v, default_sign=1),
                                         aliases=('nsc',),
                                         description='Match VALUE against time of next scrape'),
    })


class TrackerFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = _SingleFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from . import utils


class ConstantBase():
    pass

def is_constant(obj):
    return isinstance(obj, ConstantBase)

_constants_cache = {}
def get_constant(name, repr=None, bases=(), init_value=None, attrs={}):
    """
    Get constant/singleton defined by `name`

    repr: Same as `name.upper()` unless specified otherwise
    bases: tuple of base classes; will be prepended to (ConstantBase, str)
    init_value: Argument for the new constant class during creation
    attrs: Additional class attributes
    """
    if name not in _constants_cache:
        def __str__(self): return self.name
        def __repr__(self): return ('<Constant: %s>' % self.name.upper()) if self.repr is None else self.repr
        cls_attrs = {'__str__': __str__,
                     '__repr__': __repr__,
                     'name': name,
                     'repr': repr}
        cls_attrs.update(attrs)
        cls = type('Constant', bases + (ConstantBase,), cls_attrs)

        if init_value is not None:
            _constants_cache[name] = cls(init_value)
        else:
            _constants_cache[name] = cls()
    return _constants_cache[name]

DISCONNECTED = get_constant('disconnected', repr='<disconnected>')
UNLIMITED = get_constant('unlimited', bases=(utils.Float,), init_value='inf')
MAX_TORRENT_FILE_SIZE = utils.SizeInBytes(10e6)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

# flake8: noqa

# Prevent circular import
from .constants import *  # isort:skip
from .errors import *     # isort:skip

from .aiotransmission.api_settings import SettingsAPI
from .aiotransmission.api_status import StatusAPI
from .aiotransmission.api_torrent import TorrentAPI
from .aiotransmission.rpc import TransmissionRPC
from .aiotransmission.torrent import Torrent
from .api import API
from .filters import FileFilter, PeerFilter, SettingFilter, TorrentFilter, TrackerFilter
from .poll import RequestPoller
from .sorters import PeerSorter, SettingSorter, TorrentSorter, TrackerSorter
from .trequestpool import TorrentRequestPool
from .ttypes import TorrentFile, TorrentPeer, TorrentTracker
from .utils import URL, Response



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from .base import SorterBase, SortSpec


class _SortSpec(SortSpec):
    def __init__(self, *args, description='', needed_keys=(), **kwargs):
        description = 'Sort torrents by %s' % description
        super().__init__(*args, description=description, **kwargs)
        self.needed_keys = needed_keys


class TorrentSorter(SorterBase):
    DEFAULT_SORT = 'name'
    SORTSPECS = {
        'id':                _SortSpec(lambda t: t['id'],
                                       needed_keys=('id',),
                                       description='ID'),
        'name':              _SortSpec(lambda t: t['name'].casefold(),
                                       aliases=('n',),
                                       needed_keys=('name',),
                                       description='name'),
        'path':              _SortSpec(lambda t: t['path'],
                                       aliases=('dir',),
                                       needed_keys=('path',),
                                       description='download path'),
        'status':            _SortSpec(lambda t: t['status'],
                                       aliases=('st',),
                                       needed_keys=('status',),
                                       description='current status (idle, uploading, verifying, etc.)'),
        'error':             _SortSpec(lambda t: t['error'],
                                       aliases=('err',),
                                       needed_keys=('error',),
                                       description='error message'),
        'uploaded':          _SortSpec(lambda t: t['size-uploaded'],
                                       aliases=('up',),
                                       needed_keys=('size-uploaded',),
                                       description='number of uploaded bytes'),
        'downloaded':        _SortSpec(lambda t: t['size-downloaded'],
                                       aliases=('dn',),
                                       needed_keys=('size-downloaded',),
                                       description='number of downloaded bytes'),
        '%downloaded':       _SortSpec(lambda t: t['%downloaded'],
                                       lambda t: t['%metadata'],
                                       lambda t: t['%verified'],
                                       aliases=('%dn',),
                                       needed_keys=('%downloaded', '%metadata', '%verified'),
                                       description='downloading or verifying progress'),
        'size':              _SortSpec(lambda t: t['size-final'],
                                       aliases=('sz',),
                                       needed_keys=('size-final',),
                                       description='number of bytes of all wanted files'),
        'peers':             _SortSpec(lambda t: t['peers-connected'],
                                       aliases=('prs',),
                                       needed_keys=('peers-connected',),
                                       description='connected peers'),
        'seeds':             _SortSpec(lambda t: t['peers-seeding'],
                                       aliases=('sds',),
                                       needed_keys=('peers-seeding',),
                                       description='highest number of seeds reported by any tracker'),
        'ratio':             _SortSpec(lambda t: t['ratio'],
                                       aliases=('rto',),
                                       needed_keys=('ratio',),
                                       description='upload/download ratio'),
        'rate-up':           _SortSpec(lambda t: t['rate-up'],
                                       aliases=('rup',),
                                       needed_keys=('rate-up',),
                                       description='upload rate'),
        'rate-down':         _SortSpec(lambda t: t['rate-down'],
                                       aliases=('rdn',),
                                       needed_keys=('rate-down',),
                                       description='download rate'),
        'rate':              _SortSpec(lambda t: t['rate-up'] + t['rate-down'],
                                       aliases=('r',),
                                       needed_keys=('rate-up', 'rate-down'),
                                       description='combined download and upload rate'),
        'limit-rate-up':     _SortSpec(lambda t: t['limit-rate-up'],
                                       aliases=('lrup',),
                                       needed_keys=('limit-rate-up',),
                                       description='upload rate limit'),
        'limit-rate-down':   _SortSpec(lambda t: t['limit-rate-down'],
                                       aliases=('lrdn',),
                                       needed_keys=('limit-rate-down',),
                                       description='download rate limit'),
        'limit-rate':        _SortSpec(lambda t: t['limit-rate-up'] + t['limit-rate-down'],
                                       aliases=('lr',),
                                       needed_keys=('limit-rate-up', 'limit-rate-down'),
                                       description='combined download and upload rate limit'),
        'tracker':           _SortSpec(lambda t: t['trackers'][0]['url-announce'].domain if t['trackers'] else '',
                                       aliases=('trk',),
                                       needed_keys=('trackers',),
                                       description='domain of first tracker'),
        'eta':               _SortSpec(lambda t: t['timespan-eta'],
                                       needed_keys=('timespan-eta',),
                                       description='estimated time to finish downloading'),
        'created':           _SortSpec(lambda t: t['time-created'],
                                       aliases=('tcrt',),
                                       needed_keys=('time-created',),
                                       description='creation time'),
        'added':             _SortSpec(lambda t: t['time-added'],
                                       aliases=('tadd',),
                                       needed_keys=('time-added',),
                                       description='time of addition'),
        'started':           _SortSpec(lambda t: t['time-started'],
                                       aliases=('tsta',),
                                       needed_keys=('time-started',),
                                       description='start time'),
        'activity':          _SortSpec(lambda t: t['time-activity'],
                                       aliases=('tact',),
                                       needed_keys=('time-activity',),
                                       description='time of latest upload/download activity'),
        'completed':         _SortSpec(lambda t: t['time-completed'],
                                       aliases=('tcmp',),
                                       needed_keys=('time-completed',),
                                       description='time of completion'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Torrent keys we need for all sortspecs
        self._needed_keys = tuple(set().union(
            *(sortspec.needed_keys
              for sortspec in self._sortspecs)
        ))

    @property
    def needed_keys(self):
        return self._needed_keys
